"""

This script extracts various DRS performance metrics from K/L vpxd verboselog. 
Here are the things you need in order to run it:
1. A directory with vpxd logs with read permission
2. Another directory with write permission
3. matplotlib library at http://matplotlib.sourceforge.net/
4. this drslog.py binary
5. If run on windows, equivalence of "ls" "cat" "egrep" from toolchain or cygwin
6. python 2.5 or 2.6
 
After matplotlib is installed, you can do ">import matplotlib" from a python interactive shell to verify.
The command to run drslog.py is
 
\toolchain\python.exe -I\logs -O.\drslog -p stats  (Windows)
/toolchain/python -I/logs -O./drslog -p stats (Linux)
 
The output directory should be cleaned before being reused for another run (because temp files will be picked up otherwise), this can be done either manually or by 
\toolchain\python.exe -O.\drslog clean
 
[Notes]
* "-p" means plot, without "-p" it only produces one .txt file (no need for matplotlib)
* the default input directory is "."
* the default output directory is "./drslog"
* output directory can't be just "." 
* the script takes a long time for large number of logs, took more than 15 mins for 6000 logs
* use "-h" to see all arguments

[Changes]
* Guoqiang coded the first version
* Shankari added code to read DrmDump

"""


import sys
import os
import os.path
import shutil
import commands
import time
import copy
import math
from subprocess import *
from optparse import OptionParser

### Do not modify ###
ALL_COMMANDS = ["init", "alldrs", "clean", "stats"]
COMMAND_HANDLER_NAME_PREFIX = "command_handler_"

### Number of cluster with most invocations to plot ###
PLOT_TOP_CLUSTERS = 3
### Number of cluster with lowest success rate to plot ###
PLOT_BAD_CLUSTERS = 2
### Number of cluster with longest average invocation time ###
PLOT_SLOW_CLUSTERS = 2
### Ignore clusters that have few invocations ###
DUMP_MIN_INVOCATIONS = 10
DUMP_SHOW_LASTINVO = False

ROW_WIDTH = 120
ROW_BREAKER_SINGLE = "-" * ROW_WIDTH
ROW_BREAKER_DOUBLE = "=" * ROW_WIDTH

PATTERNS_ALLDRS = set([])
PATTERNS_KEYDRS = set([])


"""
Definition of invocation profiles (IP)
Each IP is reported individually
"""
IP_Common = {
   "property": {
      "DRMDUMP":"Drmdump written",
      "ALGORITHM": "Enter DRS_PROFILE Calling",
      "ALGODONE": "Exit DRS_PROFILE CallingProposeActions",
      "CCDONE": "Exit DRS_PROFILE ClusterComputeCompatibilitySet",
      "ALLOCATOR_ENTER": "Enter DRS_PROFILE VpxdDrmInstanceAllocator",
      "ALLOCATOR_EXIT": "Exit DRS_PROFILE VpxdDrmInstanceAllocator",
      "CCDIRTY": "[ClusterMo::ComputeCompatibilitySet] Dirty hosts"
   },
   "fate":  {
      "ALLOCATOR":"Another instance is already",
      "TIMEDOUT":"Cannot retrieve domain configuration",
      "EXCEPTION":"threw a"
   }
}

IP_Async = {
   "name": "async",
   "start": "AskRefreshDrmRecLro::StartWork",
   "property": {
   },
   "fate":  {
      "SUCCEED": "Done querying DRM",
      "ALGOFAIL": "Invocation to DRS algorithm failed"
   },
   "sample":["TotalTime", "SuccTime","SnapTime", "LockTime", "SuccIntv", \
   "PostAlgoTime", "NumCompat", "CompatTime", "DirtyHosts", "DirtyVMs"]
}

IP_Sync1 = {
   "name": "vm_poweron",
   "start":"Enter DRS_PROFILE GetInitialPlacementHosts",
   "property": {
   },
   "fate": {
      "SUCCEED": "Exit DRS_PROFILE GetInitialPlacementHosts",
      "EXCEPTION2":"DRS failed to recommend host for vm"
   },
   "sample": ["TotalTime", "SuccTime","SnapTime", "LockTime", "NumCompat", \
   "CompatTime", "DirtyHosts", "DirtyVMs"]
}

IP_Sync2 = {
   "name": "group_poweron",
   "start":"Enter DRS_PROFILE GetPlacementRecommendationsForPowerOn",
   "property": {
   },
   "fate": {
      "SUCCEED": "Exit DRS_PROFILE GetPlacementRecommendationsForPowerOn "
   },
   "sample": ["TotalTime", "SuccTime","SnapTime"]
}

IP_Sync3 = {
   "name": "vm_res",
   "start":"Enter DRS_PROFILE GetCachedVmInitialResSettings",
   "property": {
   },
   "fate": {
      "SUCCEED": "Exit DRS_PROFILE GetCachedVmInitialResSettings",
      "BADSTATE": "error 'App'] [VpxdDrmInterface::GetCachedVmInitialResSettings]"
   },
   "sample": ["TotalTime", "SuccTime","SnapTime", "PostAlgoTime"]
}

#### each item in the list is [name, unit, description] ###
sample_list = [
   ["TotalTime", "sec", "Duration of an invocation whose result is known"],
   ["SuccTime", "sec", "Duration of a successful invocation"],
   ["LastInvo", "", "Start time of the last invocation"],
   ["LastSucc", "", "Finish time of the last successful invocation"],
   ["LockTime", "sec", "Time spent waiting for InstanceAllocator mutex"],
   ["SnapTime", "sec", "Time between LockTime and call to algorithm"],
   ["InvoIntv", "sec", "Time between the start time of two invocations"],
   ["SuccIntv", "sec", "Time between the finish time of two successful invocation"],
   ["NumCompat", "", "Number of invocations to MoCluster::ComputeCompatibilitySet \
(non-allocOnly calls)"],
   ["CompatTime", "sec", "Time spent for compatibility test (non-allocOnly calls)"],
   ["PostAlgoTime", "sec", "Time after algorithm call finished"],
   ["DirtyHosts", "", "Number of dirty hosts recomputed"],
   ["DirtyVMs", "", "Number of dirty vm recomputed"]
]

options = {}
cluster_stats = [] # list(dict(cluster, Stats))
thread_invos = {}

### What IP we care about ###
all_profiles = [IP_Async, IP_Sync1, IP_Sync3]

all_invocations = [] # list(list(invocations))
earliest_time = None
sample_map = {}
system_stats = None
cfg = None

### System wide configurations ###
class SystemConfig:
   def __init__(self, options):
      # global variables
      self.DIR_OUTPUT = options.outputdir or "drslog"
      self.DIR_WORKING = self.DIR_OUTPUT
      self.DIR_LOG = self.DIR_OUTPUT
      self.DIR_INPUT = options.inputdir or "."
      # File that contains the list of vpxd logs, one per line
      self.FN_LIST_DEFAULT = self.DIR_OUTPUT + "/" + "logs.list"
      self.FN_PATTERN_ALLDRS = self.DIR_WORKING + "/" + "alldrs.pattern"
      self.FN_PATTERN_KEYDRS = self.DIR_WORKING + "/" + "keydrs.pattern"
      self.FN_LOG_ALLDRS = self.DIR_LOG + "/" + "alldrs.log"
      self.FN_LOG_KEYDRS = self.DIR_LOG + "/" + "keydrs.log"
      self.FN_LOG_RESULT = self.DIR_LOG + "/" + "result.log"
      self.FN_LOG_DETAIL = self.DIR_LOG + "/" + "detail.log"
      self.FN_STATS = self.DIR_WORKING + "/" + "stats.txt"
      self.FN_DIR_DRMDUMP = "."
      self.FN_PLOT = self.DIR_LOG + "/drslog"
      self.OPTION_GZ_ONLY = True
      self.SHELL_CMD_MAKELIST = "ls -1rt %s/ |"\
         "egrep -E \"vpxd-[0-9]{1,}%s\" > " \
         % (self.DIR_INPUT, (".log.gz" if self.OPTION_GZ_ONLY else ""))\
          + self.FN_LIST_DEFAULT
      
      self.SHELL_CMD_KEYDRS = "cat " + self.FN_LOG_ALLDRS + "|egrep -f " + \
         self.FN_PATTERN_KEYDRS + " > " + self.FN_LOG_KEYDRS
      self.SHELL_CMD_CAT_KEYDRS = "cat " + self.FN_LOG_KEYDRS
      self.SHELL_CMD_CAT_ALLDRS = "cat " + self.FN_LOG_ALLDRS
      self.PATTERN_FILES = { self.FN_PATTERN_ALLDRS: PATTERNS_ALLDRS, \
         self.FN_PATTERN_KEYDRS: PATTERNS_KEYDRS }
      

### Global stats ###
class SystemStats:
   def __init__(self):
      self.lock_times = {} # dict(cname, dict(lock_type, metric), list((int, int)))
      self.allocator_waitlist = {} # dict(cname, list(int,int))
      self.allocator_raw = {} # dict(cname, list(int,bool))
      return
   def add_cluster_allocator_record(self,cname, time, is_try):
      if not self.allocator_raw.has_key(cname):
         self.allocator_raw[cname] = []
      self.allocator_raw[cname].append((time, is_try))
      return
   def get_cluster_waitlist(self, cname):
      if not self.allocator_waitlist.has_key(cname):
         return None
      return self.allocator_waitlist[cname]   
   # aggregate lock/unlock requests to size of waitlist
   def calculate_cluster_waitlist(self):
      for cname, rawlist in self.allocator_raw.iteritems():
         rawlist.sort()
         self.allocator_waitlist[cname] = []
         last_time = 0
         nwait = 0
         for t, r in rawlist:
            if t!=last_time:
               self.allocator_waitlist[cname].append((t, nwait))
               last_time = t
            nwait += (1 if r else -1)
      return
   def add_cluster_lock_time(self, cname, time, lock_type, metric, ms):
      if not self.lock_times.has_key(cname):
         self.lock_times[cname] = {}
      cstats = self.lock_times[cname]
      key = (lock_type, metric)
      if not cstats.has_key(key):
         cstats[key] = []
      cstats[key].append((time, float(ms)/1000))
      return
   def get_cluster_lock_time(self, cname, lock_type, metric):
      return self.lock_times[(cname, lock_type, metric)]
   def get_all_lock_time(self, cname):
      if self.lock_times.has_key(cname):
         return self.lock_times[cname]
      else: return {}
      
class Stats:
   def __init__(self, prof):
      self.profile = prof
      self.invoCount = 0
      self.fateCountMap = {}
      for x in prof["fate"]:
         self.fateCountMap[x]=0
      self.fateCountMap["UNKNOWN"] = 0
      self.algoCount = 0
      self.succTimes = []
      self.invoTimes = []
      self.lastSuccTime = 0
      self.lastInvoTime = 0
      self.samples = {}
      for sam in self.profile["sample"]:
         self.samples[sam] = {"min":99999,"max":-1, "count":0, "total":0}
      # default samples if any
      return
   
   def printSample(self, sam):
      # format = "%.1f" if sam["min"]<10 else "%d"
      format = "%d"
      min_str = "X" if sam["min"] == 99999 else format % sam["min"]
      max_str = "X" if sam["max"] == -1 else format % sam["max"]
      avg_str = "X" if sam["count"] == 0 else format % (sam["total"]/sam["count"])
      return "%s:%s:%s" %(min_str,max_str,avg_str)
   
   def addSample(self, key, value):
      if self.samples.has_key(key):
         self.samples[key]["total"]+=value
         self.samples[key]["count"]+=1
         self.samples[key]["min"] = min(self.samples[key]["min"], value)
         self.samples[key]["max"] = max(self.samples[key]["max"], value)
      return
   
   def addIntvSamples(self):
      self.succTimes.sort()
      self.invoTimes.sort()
      self.lastSuccTime = 0 if len(self.succTimes) == 0 else self.succTimes[-1]
      self.lastInvoTime = 0 if len(self.invoTimes) ==0  else self.invoTimes[-1]
      for x in range(1,len(self.succTimes)):
         self.addSample("SuccIntv", self.succTimes[x] - self.succTimes[x-1])
      for x in range(1,len(self.invoTimes)):
         self.addSample("InvoIntv", self.invoTimes[x] - self.invoTimes[x-1])
      return

   def addInvocation(self,invo,outfile):
      global all_invocations
      global earliest_time
      if invo.startTime is None:
         LOGV("bad invo")
         return
      if not invo.lockTime:
         LOGV("An invocation has no lockTime: cluster = %s" % invo.encode())
         invo.tryLockTime = invo.lockTime = invo.startTime
      if not earliest_time:
         earliest_time = invo.startTime
      else: 
         earliest_time = min(earliest_time, invo.startTime)
      if options.cname and options.cname!=invo.cname:
         LOGV("Invocation of unwanted cluster %s" % invo.cname)
         return
      LOGV("Adding invocation %s %s %s %s %s %s, %s, %s" % \
         (invo.cname, invo.startTime, invo.algoTime, \
         invo.endTime, invo.profile["name"], invo.fate,\
         invo.n_hosts,invo.n_vms
         ))
      all_invocations[invo.profile["id"]].append(copy.deepcopy(invo))
      self.invoCount+=1
      if invo.lockTime:
         self.addSample("LockTime", invo.lockTime - invo.startTime)
         system_stats.add_cluster_allocator_record(invo.cname, invo.lockTime, False)
         system_stats.add_cluster_allocator_record(invo.cname, invo.tryLockTime, True)
      if invo.algoTime:
         self.algoCount+=1
         self.addSample("SnapTime", invo.algoTime - invo.lockTime)
      if invo.fate != "UNKNOWN":
         if invo.endTime is None:
            LOGV("bad invo")
            return
         self.addSample("TotalTime", invo.endTime - invo.lockTime)
         if invo.algoDoneTime:
            self.addSample("PostAlgoTime", invo.endTime - invo.algoTime)
      self.fateCountMap[invo.fate]+=1
      if invo.fate == "SUCCEED":
         self.addSample("SuccTime", invo.endTime - invo.lockTime)
         self.succTimes.append(invo.endTime)
      if invo.n_cc > 0: # alloc_only invo will have n_cc==0
         self.addSample("NumCompat", invo.n_cc)
         self.addSample("CompatTime", invo.ccTime/1000)
         self.addSample("DirtyHosts", invo.dirty_hosts)
         self.addSample("DirtyVMs", invo.dirty_vms)
      self.invoTimes.append(invo.startTime)
      str = invo.cname
      str += ",%s" % invo.profile["name"]
      str += ",%s" % make_time(invo.startTime, date = True)
      str += "," if invo.n_cc ==0 else ",%d" %(invo.ccTime/1000)
      str += "," if not invo.algoTime else ",%d" %(invo.algoTime - invo.lockTime)
      str += "," if not invo.endTime else ",%d" % (invo.endTime - invo.lockTime)
      str += ",%s" % invo.fate
      str += ",%s" % invo.thread
      str += ",%s" % invo.n_hosts
      str += ",%s" % invo.n_vms
      str += "," if invo.n_cc ==0 else ",%d" % (invo.n_cc)
      str += "," if invo.n_cc ==0 else ",%s" % (invo.ccBreakDown)
      str += ",%s" % invo.drmdump

      outfile and outfile.write("%s\n" % str)
      return

class Invocation:
   def __init__(self):
      self.reset()
      return
   def encode(self):
      return "[cname=%s, thread=%s, profile=%s, fate=%s]" % \
         (self.cname, self.thread,self.profile["name"], self.fate)
   def reset(self):
      self.cname = "UNKNOWN"
      self.startTime = None
      self.tryLockTime = None # shen instance allocation is tried
      self.lockTime = None # when instance allocator is acquired
      self.endTime = None
      self.fate = "UNKNOWN"
      self.algoTime = None
      self.algoDoneTime = None
      self.thread = None
      self.n_vms = 0
      self.n_hosts = 0
      self.dirty_vms = 0
      self.dirty_hosts = 0
      self.profile = None
      self.n_cc = 0
      self.ccTime = 0
      self.ccBreakDown = ""
      self.drmdump = "UNKNOWN"

def LOGV(message):
   if options.verbose:
      print("##VER## %s " % message)
def LOG(message):
   print("##LOG## %s " % message)
def LOGERR(message):
   print("##ERR## %s " % message)
def PANIC(message="internal error"):
   print("#PANIC# %s " % message)
   sys.exit()

def init_config():
   global cfg
   cfg = SystemConfig(options)
   
def init_patterns():
   global PATTERNS_KEYDRS
   global PATTERNS_KEYDRS
   PATTERNS_KEYDRS_ = set([])
   PATTERNS_ALLDRS_ = set([])
   PATTERNS_ALLDRS_.update(["DRS_PROFILE", "[VpxdDrmInterface", "Drmdump written"])
   PATTERNS_ALLDRS_.update(["'MoLock']"])
   PATTERNS_ALLDRS_.update(["[ClusterMo::ComputeCompatibilitySet"])
   
   for profile in all_profiles:
      PATTERNS_KEYDRS_.add(profile["start"])
      PATTERNS_KEYDRS_.update(profile["property"].values())
      PATTERNS_KEYDRS_.update(profile["fate"].values())
   
   PATTERNS_KEYDRS_.add("'MoLock'] Locking vim.ClusterComputeResource:")
   PATTERNS_KEYDRS_.add("'MoLock'] ***WARNING*** Lock domain")
   # escape special chars
   lambda_sub = lambda s: s.replace("*", "\\*").replace("[","\\[").replace("]", "\\]")
   for x in PATTERNS_ALLDRS_: PATTERNS_ALLDRS.add(lambda_sub(x))
   for x in PATTERNS_KEYDRS_: PATTERNS_KEYDRS.add(lambda_sub(x))
   return
   
def init_stats():
   global cluster_stats
   global all_invocations
   global sample_map
   global system_stats
   system_stats = SystemStats()
   for (i, p) in enumerate(all_profiles):
      p["id"] = i
      # apply IP_Common 
      tmp = p["property"]
      p["property"].update(IP_Common["property"])
      p["property"].update(tmp)
      tmp = p["fate"]
      p["fate"].update(IP_Common["fate"])
      p["fate"].update(tmp)
      LOGV("\r\n".join(["========== Patterns ==========", str(p)]))
      cluster_stats.append({})
      cluster_stats[i]["UNKNOWN"] = Stats(p)
      all_invocations.append([])
   for sam in sample_list:
      sample_map[sam[0]] = {"unit":sam[1], "desp":sam[2]}


def write_list_to_file(fullname, l, append = False):
   LOGV("writing %s to %s" % (l, fullname))
   f = open(fullname, 'a' if append else 'w' )
   for str in l:
      f.write(str+"\n")
   f.close()
   return

# execute command and return output
# use subprocess to be platform independent
def execute_shell_command(command):
   LOGV("executing: %s" % command)
   subp = Popen(command, shell=True, stdout=PIPE)
   return subp.communicate()[0]

# get epoch seconds from a line
# format: [2008-09-18 00:37:16.500 05640 ....
def get_time(line):
   index_start = line.index("[") + 1
   index_end = line.index(".")
   try:
      t1 = time.strptime(line[index_start:index_end], "20%y-%m-%d %H:%M:%S")
   except:
      LOGERR("Bad line format: %s" %line)
      return None
   return time.mktime(t1)

# convert from epoch second to string
def make_time(t, date=False):
   if t ==0: return None
   try:
      if date: return time.strftime("%m-%d-%H:%M",time.localtime(t))
      else: return time.strftime("%H:%M:%S",time.localtime(t))
   except:
      LOGERR(t)
      return ""

# get number of host and vm from drmdump, return (host_list,vm_list)
def get_drmdump_info(line):
   if not options.drmdump:
      return (0,0,"")
   index = line.index("/drmdump")
   fdump = cfg.FN_DIR_DRMDUMP+ line[index:].rstrip("\r\n") #\r\n in the end
   fdump += ".gz"
   if not os.path.exists(fdump):
      LOGERR("Cannot find drmdump file %s" % fdump)
      return (0,0,"UNKNOWN")
   cmd_host = "cat " + fdump + " |gunzip|grep DrmHost"
   hosts = execute_shell_command(cmd_host).split()
   cmd_vm = "cat " + fdump + " |gunzip|grep DrmNode"
   vms = execute_shell_command(cmd_vm).split()
   print hosts, vms, fdump
   return (len(hosts), len(vms), fdump)

# return the profile if the line is the start of an invocation
def get_new_invocation(line):
   for profile in all_profiles:
      if line.find(profile["start"]) > -1:
         return profile
   return None

# dumping functions
def dump_sample_description():
   print ROW_BREAKER_DOUBLE
   for name, item in sample_map.iteritems():
      s = name
      if len(item["unit"]) >0: s+= "(" + item["unit"] + ")"
      s+= ": " + item["desp"]
      print s
   print ROW_BREAKER_DOUBLE
   return
   
def dump_stats(pid):
   lines = []
   def print_row(items, width, is_border = False):
      l = items[0][0] if is_border else "| "
      for i, x in enumerate(items):
         l += str(x)
         padding = x[0] if is_border else ' '
         l += max(0,width[i] - len(str(x)))* padding
         if not is_border: l += " | "
         else: l += (padding * 3)
      print l
      lines.append(l)
      return
   
   LOG("Dumping stats for %s..." % all_profiles[pid]["name"])
   lines.append("\n\nDumping stats for %s..." % all_profiles[pid]["name"])
   profile = all_profiles[pid]
   pstats = cluster_stats[pid]
   
   ### print numbers
   items = ["CLUSTER","INVOCATION","ALGOCALL"]
   fates = set(profile["fate"].keys())
   for fate in fates:
      items.append(fate)
   print_row(["="] * len(items), [12] * len(items), is_border = True)
   print_row(items,[12]* len(items))
   print_row([("-" * len(x)) for x in items], [12] * len(items))
   sorted_cluster = pstats.keys()
   comp_invo = lambda c2,c1: cmp(pstats[c1].invoCount, pstats[c2].invoCount)
   sorted_cluster.sort(comp_invo)
   for c in sorted_cluster:
      stat = pstats[c]
      if stat.invoCount < DUMP_MIN_INVOCATIONS: continue
      items = []
      items.append(c[7:] if c!="UNKNOWN" else "UNKNOWN")
      items.append(stat.invoCount)
      items.append("%d(%%%d)" %(stat.algoCount, 100*stat.algoCount/stat.invoCount))
      for fate in fates:
         items.append("%d(%%%d)" %(stat.fateCountMap[fate], \
         100*stat.fateCountMap[fate]/stat.invoCount))
      print_row(items, [12]* len(items))
   print_row(["="] * len(items), [12] * len(items), is_border = True)
   print
   
   ### print stats
   sample_keys = profile["sample"]
   items = ["CLUSTER"]
   if DUMP_SHOW_LASTINVO: items.extend(["LastInvo", "LastSucc"])
   for sk in sample_keys:
      items.append(sk + "[" + sample_map[sk]["unit"] + "]")
   widths = [max(len(x),11) for x in items]
   print_row([ ("=" * len(x)) for x in items], widths, is_border = True)
   print_row(items, widths)
   print_row([ ("-" * len(x)) for x in items], widths)
   items = ["","",""] if DUMP_SHOW_LASTINVO else [""]
   items += ["min:max:avg"] * len(sample_keys)
   print_row(items,widths)
   for c in sorted_cluster:
      stat = pstats[c]
      if stat.invoCount < DUMP_MIN_INVOCATIONS: continue
      items = []
      items.append(c[7:] if c!="UNKNOWN" else "UNKNOWN")
      if DUMP_SHOW_LASTINVO:
         items.append(make_time(stat.lastInvoTime) or "X")
         items.append(make_time(stat.lastSuccTime) or "X")
      for sk in sample_keys:
         s = stat.printSample(stat.samples[sk])
         items.append(s)
      print_row(items,widths)
   print_row([ ("=" * x) for x in widths], widths, is_border = True)
   print
   
   write_list_to_file(cfg.FN_STATS, lines, append = True)
   return

# post_process stats
def post_process_stats():
   global system_stats
   LOG("Post processing stats")
   system_stats.calculate_cluster_waitlist()
   return

# update invocations based on type
def update_async(time, line, invo, isStart = False):
   # find cluster name
   if invo.cname == "UNKNOWN":
      ind = line.find("Asking DRM for recommendations on cluster domain")
      if ind > -1:
         cname = line[line.index("domain"):].rstrip()
         invo.cname = cname
         pstats = cluster_stats[invo.profile["id"]]
         if not pstats.has_key(cname): 
            pstats[cname]= Stats(invo.profile)
         return
   if line.find(invo.profile["property"]["CCDONE"]) > -1:
      invo.n_cc += 1
      ind1 = line.rfind("(") + 1
      ind2 = line.find(" ", ind1)
      invo.ccTime += int(line[ind1:ind2])
      invo.ccBreakDown +=" " + line[ind1:ind2]
      return
   return

# update invocations based on type
def update_sync1(time, line, invo, isStart = False):
   if line.find(invo.profile["property"]["CCDONE"]) > -1:
      invo.n_cc += 1
      ind1 = line.rfind("(") + 1
      ind2 = line.find(" ", ind1)
      invo.ccTime += int(line[ind1:ind2])
      invo.ccBreakDown +=" " + line[ind1:ind2]
      return
   # drmdump
   if line.find(invo.profile["property"]["DRMDUMP"]) > -1:
      ind = line.index("cluster")
      ind2 = line.index("/", ind)
      invo.cname = "domain-c" + line[ind+7:ind2]
      pstats = cluster_stats[invo.profile["id"]]
      if not pstats.has_key(invo.cname):
         pstats[invo.cname]= Stats(invo.profile)
   return

# update invocations based on type
def update_sync3(time, line, invo, isStart = False):
   # drmdump
   if line.find(invo.profile["property"]["DRMDUMP"]) > -1:
      ind = line.index("cluster")
      ind2 = line.index("/", ind)
      invo.cname = "domain-c" + line[ind+7:ind2]
      pstats = cluster_stats[invo.profile["id"]]
      if not pstats.has_key(invo.cname):
         pstats[invo.cname]= Stats(invo.profile)
   return

# update invocations based on type
def update_invocation(time, line, invo, isStart = False):
   # common properties
   if line.find(invo.profile["property"]["ALGORITHM"]) > -1:
      invo.algoTime = time
   elif line.find(invo.profile["property"]["ALGODONE"]) > -1:
      invo.algoDoneTime = time
   elif line.find(invo.profile["property"]["ALLOCATOR_EXIT"]) > -1:
      invo.lockTime = time
   elif line.find(invo.profile["property"]["ALLOCATOR_ENTER"]) > -1:
      invo.tryLockTime = time
   elif line.find(invo.profile["property"]["CCDIRTY"]) > -1:
      segs = line.split(" ")
      invo.dirty_hosts += int(segs[8])
      invo.dirty_vms += int(segs[12][:-1])
      #print "dirty: ", dh, dvm
   elif line.find(invo.profile["property"]["DRMDUMP"]) > -1:
      (nh,nv,fdump) = get_drmdump_info(line)
      invo.n_hosts = int(nh)
      invo.n_vms = int(nv)
      invo.drmdump = fdump

      
   # profile specific properties
   if invo.profile == IP_Async: 
      update_async(time, line, invo, isStart)
   elif invo.profile == IP_Sync1:
      update_sync1(time, line, invo, isStart)
   elif invo.profile == IP_Sync3:
      update_sync3(time, line, invo, isStart)
   else: return
   # sort the fate list so prefix is handled first
   for fate,fkey in invo.profile["fate"].iteritems():
      if line.find(fkey) > -1:
         invo.fate = fate
         invo.endTime = time
   if invo.profile == IP_Sync1 and invo.fate == "SUCCEED" and invo.algoTime is None:
      # report
      LOGERR("weird..%s" % line)
      invo.fate = "UNKNOWN"
      #PANIC()
   return

"""
[2009-01-07 10:43:51.772 10308 warning 'MoLock'] ***WARNING*** Lock host-1955 mode EXCLUSIVEALL held for 1069 ms
[2009-01-07 10:43:51.959 02596 verbose 'MoLock'] Locking vim.ClusterComputeResource:domain-c5973 in mode: SHARE took 1224 ms
return True if this line should be skipped for the rest of processing
"""
def update_lock_time(time, line):
   if line.find("'MoLock'") == -1: return False
   index1 = line.find("domain")
   if index1 == -1: return False
   cname = line[index1 : line.find(" ", index1)]
   if line.find("Locking") > -1: metric = "acquire"
   elif line.find("held") > -1: metric = "hold"
   else: return True
   index2 = line.find("mode")
   index2 = line.find(" ", index2)
   mode = line[index2 + 1 : line.find(" ", index2 + 1)]
   index3 = line.rfind(" ", 0, -5)
   ms = line[index3 + 1 : line.find(" ", index3 + 1)]
   #print cname, time, mode, metric, ms
   system_stats.add_cluster_lock_time(cname, time, mode, metric, ms)
   return True
   
def visit_keydrs(line, lineno, outfile, detailfile):
   global cluster_stats
   detail = None
   t = get_time(line)
   if not t:
      LOG("Skip line: " + line)
      return
   if update_lock_time(t, line):
      return
   col_th = 3 if options.version == "25" else 2
   th = line.split()[col_th]
   if not thread_invos.has_key(th):
      thread_invos[th] = Invocation()
   invocation = thread_invos[th]
   # is a new invocation started?
   profile = get_new_invocation(line)
   if profile:
      # add previous invocation
      if invocation.profile:
         pstats = cluster_stats[invocation.profile["id"]]
         if invocation.fate == "UNKNOWN":
            LOG("Unknown invocation: cname=%s, thread=%s ,line=%s" \
               %(invocation.encode(), invocation.thread, line))
         pstats[invocation.cname].addInvocation(invocation,outfile)
      invocation.reset()
      invocation.profile = profile
      invocation.startTime = t
      invocation.thread = th
      update_invocation(t, line, invocation, True)
   else:
      # ignore the lines with no invocation
      if (not invocation.startTime) or invocation.endTime:
         LOGV("Orphan line: " + line)
         detail = "%s %s %d [ORPH]%s" %(invocation.cname, invocation.thread, lineno, line) 
         detailfile and detailfile.write("%s" % detail)
         return
      update_invocation(t, line, invocation, False)
   detail = "%s %s %d %s" % (invocation.cname, invocation.thread, lineno, line) 
   detailfile and detailfile.write("%s" % detail)
   return

def compose_alldrs_cmd(fn):
   cmd = "cat " + fn
   if fn.endswith(".gz"): cmd += "| gunzip -d "
   cmd += " |egrep -f " + cfg.FN_PATTERN_ALLDRS + " >> " + cfg.FN_LOG_ALLDRS
   return cmd

def generate_alldrs():
   files = set([])
   if not os.path.exists(cfg.FN_LIST_DEFAULT):
      execute_shell_command(cfg.SHELL_CMD_MAKELIST)
   # remove old file
   if os.path.exists(cfg.FN_LOG_ALLDRS):
      os.remove(cfg.FN_LOG_ALLDRS)
   fin = open(cfg.FN_LIST_DEFAULT,'r')
   try:
      for line in fin:
         fn = line.lstrip().rstrip(" \r\n")
         fn = cfg.DIR_INPUT + "/" + fn
         if len(fn) == 0: continue
         to_log = "Processing %s ..." % fn
         files.add(fn)
         execute_shell_command(compose_alldrs_cmd(fn))
         to_log += "done"
         LOG(to_log)
      if os.path.exists(cfg.FN_LOG_ALLDRS):
         LOG("%s has been created successfully" % cfg.FN_LOG_ALLDRS)
      else:
         LOG("No DRS related logs have been processed")
         return False
   except IOError:
      LOGERR("Error reading file")
      return False
   finally:
      fin.close()
   return True
   
def generate_keydrs():
   # requires alldrs
   if not os.path.exists(cfg.FN_LOG_ALLDRS):
      LOGERR("%s needs to be generated first, try -h for help" % cfg.FN_LOG_ALLDRS)
      return False
   LOGV("Regenerating pattern files...")
   command_handler_init()
   LOGV("This operation could take a while...")
   execute_shell_command(cfg.SHELL_CMD_KEYDRS)
   LOG("%s has been created successfully" % cfg.FN_LOG_KEYDRS)
   return True

def analyze():
   # assume keydrs exists
   LOG("analyzing logs, this could take a while...")
   fout = None
   fdetail = None
   if options.output:
      fout = open(cfg.FN_LOG_RESULT, 'w')
      fdetail = open(cfg.FN_LOG_DETAIL, 'w')
   fin = open(cfg.FN_LOG_KEYDRS,'r')
   try:
      lineno = 0
      for line in fin:
         visit_keydrs(line, lineno, fout, fdetail)
         lineno+=1
      # add last invocations
      for invo in thread_invos.values():
         if invo.cname and invo.profile:
            if invo.fate != "UNKNOWN":
               cluster_stats[invo.profile["id"]][invo.cname].addInvocation(invo, fout)
   except IOError:
      LOGERR("Error reading file")
   finally:
      fin.close()
      fout and fout.close()
      fdetail and fdetail.close()
   fout and LOG("%s has been created successfully" % cfg.FN_LOG_RESULT)
   
   # post_process stats
   post_process_stats()
   dump_sample_description()
   write_list_to_file(cfg.FN_STATS, [])
   for pid in range(0, len(all_profiles)):
      pstats = cluster_stats[pid]
      for c,stat in pstats.iteritems():
         stat.addIntvSamples()
      LOG("dumping statistics...")
      dump_stats(pid)
      if options.plot:
         plot(pid)
   return

##########################################################################
# plotting functions 
##########################################################################

def util_plot_avg_hor(ax, values):
   avg = float(sum(values))/len(values)
   ax.axhline(y = avg, c = 'r', linewidth=2)

def util_plot_xgrid(ax, begin, end, interval):
   import matplotlib.pyplot as plt
   import matplotlib
   from pylab import arange
   # adjust interval to avoid collapsed text label
   interval = max(interval, (end-begin)/30)
   ax.xaxis.set_ticks(arange(begin, end, interval))
   plt.setp(ax.get_xticklabels(), rotation=30, fontsize=10)
   def format_date(x,pos=None):
      return make_time(earliest_time + x, date = True)
   ax.xaxis.set_major_formatter(matplotlib.ticker.FuncFormatter(format_date))
   
# finish percentage with time
def plot_all_finish(ax, pid):
   N = len(all_invocations[pid])
   total_times = [(x.endTime - x.startTime) for x in all_invocations[pid] if x.fate == "SUCCEED"]
   if len(total_times) ==0: return
   max_t = int(max(total_times))
   secs = range(0, max_t+1)
   finish = [len([x for x in total_times if x<t]) for t in secs]
   finish_perc = [ x*100/N for x in finish ]
   ax.plot(finish_perc, c = 'b', linewidth = 4)
   ax.axhline(y = 50, c = 'r', linewidth=2)
   ax.axhline(y = 90, c = 'r', linewidth=2)
   ax.set_xlabel(" "*150 + "Sec")
   ax.set_ylabel("% of Invo.", va="top")
   ax.set_title("Successful finish time for all invocations")
  # ax.set_xscale("log", basex = 10)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.grid()
   return

"""
# post algo time
def plot_all_postalgo(ax, pid):
   N = len(all_invocations[pid])
   succ_invocations = [x for x in all_invocations[pid] if \
      (x.fate == "SUCCEED" and x.algoDoneTime)]
   if len(succ_invocations) ==0: return
   time_sec = [(x.startTime - earliest_time) for x in succ_invocations]
   post_algo_times = [(x.endTime - x.algoDoneTime) for x in succ_invocations]
   size = [(16 if not x.n_hosts else x.n_hosts/2) for x in succ_invocations]
   color = [cluster_stats[pid].keys().index(x.cname) for x in succ_invocations]
   if (len(post_algo_times) >0):
      ax.scatter(time_sec, post_algo_times, s = size, c = color)
   ax.set_ylabel("Sec", va="top")
   util_plot_avg_hor(ax, post_algo_times)
   util_plot_xgrid(ax, 0, max(time_sec), 3600)
   ax.set_title("Total Time Spent After Algo Call")
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.grid(True)
   return
"""
# post algo time
def plot_cluster_postalgo(ax, cname, pid):
   N = len(all_invocations[pid])
   succ_invocations = [x for x in all_invocations[pid] if \
      (x.fate == "SUCCEED" and x.algoDoneTime and x.cname == cname)]
   if len(succ_invocations) ==0: return
   time_sec = [(x.startTime - earliest_time) for x in succ_invocations]
   post_algo_times = [(x.endTime - x.algoDoneTime) for x in succ_invocations]
   size = [(16 if not x.n_hosts else x.n_hosts/2) for x in succ_invocations]
   if (len(post_algo_times) >0):
      ax.scatter(time_sec, post_algo_times, s = size)
   ax.set_ylabel("Sec", va="top")
   util_plot_avg_hor(ax, post_algo_times)
   util_plot_xgrid(ax, 0, max(time_sec), 3600)
   ax.set_title("Total Time Spent After Algo Call for %s" % cname)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.grid(True)
   return

# instance allocator time
def plot_cluster_allocator(ax, cname, pid):
   N = len(all_invocations[pid])
   all_invo = [x for x in all_invocations[pid] if (x.cname == cname)]
   if len(all_invo) ==0: return
   time_sec = [(x.startTime - earliest_time) for x in all_invo]
   lock_times = [(x.lockTime - x.startTime) for x in all_invo]
   size = [(16 if not x.n_hosts else x.n_hosts/2) for x in all_invo]
   ax.scatter(time_sec, lock_times, s = size)
   ax.set_ylabel("Sec", va="top")
   util_plot_avg_hor(ax, lock_times)
   util_plot_xgrid(ax, 0, max(time_sec), 3600)
   ax.set_title("Time Spent to acquire DrmInstanceAllocator for %s" % cname)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.grid(True)
   return

# instance allocator wait list
def plot_cluster_allocator_wl(ax, cname, pid):
   stat = system_stats.get_cluster_waitlist(cname)
   if stat is None: return
   # print cstats
   invocations = [x for x in all_invocations[pid] if (x.cname == cname)]
   time_sec = [x.startTime - earliest_time for x in invocations]
   latest_time = max(time_sec)
   xv = [(x - earliest_time) for (x,y) in stat]
   yv = [y for (x,y) in stat]
   ax.plot(xv, yv)
   ax.set_title("Number of waiting DRS instances for cluster %s" % cname)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   util_plot_xgrid(ax, 0, latest_time, 3600)
   ax.grid(True)

# finish percentage with time
def plot_cluster_finish(ax, cname, pid):
   invocations = [x for x in all_invocations[pid] if x.cname == cname]
   N = len(invocations)
   total_times = [(x.endTime - x.lockTime) for x in invocations if x.fate == "SUCCEED"]
   if len(total_times) ==0: return
   max_t = int(max(total_times))
   secs = range(0, max_t + 1)
   finish = [len([x for x in total_times if x<t]) for t in secs]
   finish_perc = [ x*100/N for x in finish ]
   ax.plot(finish_perc, c = 'b', linewidth = 4)
   ax.axhline(y = 50, c = 'r', linewidth=2)
   ax.axhline(y = 90, c = 'r', linewidth=2)
   ax.set_xlabel(" "*150 + "Sec")
   ax.set_ylabel("Perc. of Invo.", va="top")
   ax.set_title("Successful finish time for %s" % cname)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.grid()
   return

# compat comput time for non-alloc_only calls
def plot_cluster_ccperc(ax, cname, pid):
   invocations = [x for x in all_invocations[pid]  \
      if (x.cname == cname and x.fate == "SUCCEED" and x.n_cc > 0)]
   N = len(invocations)
   time_sec = [(x.startTime - earliest_time) for x in invocations \
      if x.lockTime!=x.algoTime]
   ccperc = [ (float(x.ccTime)/1000*100/(x.algoTime-x.lockTime)) for x in invocations \
      if x.lockTime!=x.algoTime]
   if len(time_sec) ==0: return
   ax.scatter(time_sec, ccperc, c = 'b', linewidth = 2)
   ax.set_ylabel("Percent", va="top")
   # do not count totaltime=0 in avg
   util_plot_avg_hor(ax, ccperc)
   latest_time = max(time_sec)
   util_plot_xgrid(ax, 0, latest_time, 3600)
   ax.set_title("Percentage of Snapshot Time Spent on Compute Compability  %s" % cname)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0, ymax=100)
   ax.grid()
   return

# compat comput time for non-alloc_only calls
def plot_cluster_cctime(ax, cname, pid):
   invocations = [x for x in all_invocations[pid] \
      if (x.cname == cname and x.fate == "SUCCEED" and x.n_cc > 0) ]
   N = len(invocations)
   time_sec = [ (x.startTime - earliest_time) for  x in invocations]
   ncc = [x.ccTime/1000 for x in invocations]
   if len(ncc) ==0: return
   ax.scatter(time_sec, ncc, c = 'b', linewidth = 2)
   ax.set_ylabel("Sec", va="top")
   util_plot_avg_hor(ax, ncc)
   latest_time = max(time_sec)
   util_plot_xgrid(ax, 0, latest_time, 3600)
   ax.set_title("ComputeCompatibility Total Time for %s" % cname)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.grid()
   return

# compat comput time breakdown, time for each iteration
def plot_cluster_ccbreakdown(ax, cname, pid):
   from pylab import arange
   threshold = 25
   invocations = [x for x in all_invocations[pid] \
      if (x.cname == cname and x.fate == "SUCCEED" and x.ccTime/1000 > threshold) ]
   N = len(invocations)
   for inv in invocations:
      times = [(int(x)/1000) for x in inv.ccBreakDown.strip().split()]
      ax.plot(range(1, len(times)+ 1), times)
      # print times
   ax.xaxis.set_ticks(arange(1, 10, 1))
   ax.set_title("Long CC Time (>%s sec total) Break-down for %s" % (threshold,cname))
   ax.set_xlim(xmin = 1)
   ax.set_ylim(ymin = 0)
   ax.set_xlabel(" "*150 + "#Iteration")
   ax.set_ylabel("Sec", va="top")
   ax.grid()
   return


# compat comput time 
def plot_cluster_ncc(ax, cname, pid):
   invocations = [x for x in all_invocations[pid]\
      if (x.cname == cname and x.fate == "SUCCEED" and x.n_cc > 0)]
   N = len(invocations)
   ncc = [x.n_cc for x in invocations]
   if len(ncc) ==0: return
   max_t = int(max(ncc))
   counts = range(0, max_t + 1)
   """
   finish = [len([x for x in ncc if x<=t]) for t in counts]
   finish_perc = [ x*100/N for x in finish ]
   ax.plot(finish_perc, c = 'b', linewidth = 4)
   ax.axhline(y = 50, c = 'r', linewidth=2)
   ax.axhline(y = 90, c = 'r', linewidth=2)
   """
   
   finish = [(len([x for x in ncc if x==t])*100)/N for t in counts]
   ax.bar(counts, finish)
   
   ax.set_xlabel(" "*150 + "Iterations")
   ax.set_ylabel("Perc. of Invo.", va="top")
   ax.set_title("Iterations of ComputeCompatibility per Invocation for %s" % cname)
   ax.set_xlim(xmin = 0)
   ax.set_xlim(xmax = (max(10, max_t)))
   ax.set_ylim(ymin = 0)
   ax.grid()
   return

# scatter graph of all invocations of a single cluster
def plot_cluster_scatter(ax, cname, pid):
   invocations = [x for x in all_invocations[pid] if (x.cname == cname and x.fate == "SUCCEED")]
   series = 0
   latest_time = 0
   if len(invocations) >0:
      time_sec = [(x.startTime - earliest_time) for x in invocations]
      time_total = [(x.endTime - x.lockTime) for x in invocations]
      size = [(32 if not x.n_hosts else x.n_hosts/2) for x in invocations]
      series += 1
      ax.scatter(time_sec, time_total, s = size, c = 'g', marker = "o")
      util_plot_avg_hor(ax, time_total)
      latest_time = max(time_sec)
      
   f_invocations = [x for x in all_invocations[pid] if (x.cname == cname and x.fate != "SUCCEED")]
   if len(f_invocations) > 0:
      f_time_sec = [ (x.startTime - earliest_time) for x in f_invocations]
      f_time_total = [ (0 if not x.endTime else (x.endTime - x.lockTime)) for x in f_invocations]
      if len(f_time_total) >0:
         ax.scatter(f_time_sec, f_time_total, s = 80, c = 'r', marker = "^")
         series += 1

   ax.set_ylabel("Sec", va="top")
   ax.set_title("Total Time of Invocations for %s" % cname)
   util_plot_xgrid(ax, 0, latest_time, 3600)
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   if series == 2:
      ax.legend(["Failed Invo.", "Succ Invo."])
   ax.grid(True)
   return

# scatter graph of success invocations
def plot_success_scatter(ax, pid):
   succ_invocations = [x for x in all_invocations[pid] if x.fate == "SUCCEED"]
   if len(succ_invocations) == 0: return
   # time_hr = [(x.startTime - earliest_time)/3600 for x in succ_invocations]
   time_sec = [x.startTime - earliest_time for x in succ_invocations]
   time_total = [(x.endTime - x.startTime) for x in succ_invocations]
   size = [(16 if not x.n_hosts else x.n_hosts/2) for x in succ_invocations]
   color = [cluster_stats[pid].keys().index(x.cname) for x in succ_invocations]
   if len(time_total) >0:
      ax.scatter(time_sec, time_total, s = size, c = color)
   ax.set_ylabel("Sec", va="top")
   util_plot_xgrid(ax, 0, time_sec[-1], 3600)
   util_plot_avg_hor(ax, time_total)
   ax.set_title("Total Time of Successful Invocations")
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.grid(True)
   return

def plot_cluster_lock(ax, cname, pid):
   # print system_stats.lock_times
   cstats = system_stats.get_all_lock_time(cname)
   if cstats is None: return
   # print cstats
   invocations = [x for x in all_invocations[pid] if (x.cname == cname)]
   time_sec = [x.startTime - earliest_time for x in invocations]
   if len(time_sec) ==0: return
   latest_time = max(time_sec)
   legends = []
   markers = ['o', 's', 'x', '+', '^','p']
   colors  = ['b','r','y','g']
   index = 0
   for (key, stat) in cstats.iteritems():
      (type, metric) = key
      if metric!="acquire": continue # ignore held lock data
      legends.append(type)
      xv = [(x - earliest_time) for (x,y) in stat]
      yv = [y for (x,y) in stat]
      ax.scatter(xv, yv, label = str(key), marker = markers[index], color = colors[index])
      index += 1
   ax.set_title("Lock acquiring delay for cluster %s" % cname)
   ax.set_ylabel("Sec", va="top")
   ax.set_xlim(xmin = 0)
   ax.set_ylim(ymin = 0)
   ax.legend(legends, loc="upper left", shadow=True)
   util_plot_xgrid(ax, 0, latest_time, 3600)
   ax.grid(True)
   return

def plot(pid):
   pname = all_profiles[pid]["name"]
   LOG("start plotting %s..." % pname)
   import matplotlib 
   import matplotlib.pyplot as plt
   import numpy
   #print matplotlib.rcParams.keys()
   pstats = cluster_stats[pid]
   comp_invo = lambda c2,c1: cmp(pstats[c1].invoCount, pstats[c2].invoCount)
   comp_fail = lambda c1,c2: cmp(\
      float(pstats[c1].fateCountMap["SUCCEED"])/pstats[c1].invoCount,\
      float(pstats[c2].fateCountMap["SUCCEED"])/pstats[c2].invoCount)
   comp_time = lambda c2,c1: cmp(\
      float(pstats[c1].samples["TotalTime"]["total"])/float(pstats[c1].samples["TotalTime"]["count"]),\
      float(pstats[c2].samples["TotalTime"]["total"])/float(pstats[c2].samples["TotalTime"]["count"]))
      
   sorted_cluster = cluster_stats[pid].keys()
   sorted_cluster = filter(lambda x:pstats[x].invoCount > 0,sorted_cluster)
      
   plotted_cluster = set([])
   sorted_cluster.sort(comp_invo)
   plotted_cluster.update(sorted_cluster[0:PLOT_TOP_CLUSTERS])
   
   sorted_cluster.sort(comp_fail)
   plotted_cluster.update(sorted_cluster[0:PLOT_BAD_CLUSTERS])
   
   sorted_cluster.sort(comp_time)
   plotted_cluster.update(sorted_cluster[0:PLOT_SLOW_CLUSTERS])
   
   n_top = len(plotted_cluster)
   if n_top ==0: return
   n_type = 10 # all invo + all time + ncc + postalgo \
              # + instance + instcctime + ccper + \
              # + ccbreakdown + lock + instance_wl
   n_col = n_top
   n_row = n_type
   plot_width = 12
   plot_height = 4
   fig = plt.figure(figsize=(plot_width * n_col, n_row * plot_height))
   fig.subplots_adjust(wspace = 0.25)
   fig.subplots_adjust(hspace = 0.5)
   fig.subplots_adjust(top = 0.95)
   fig.subplots_adjust(left = 0.05)
   fig.subplots_adjust(bottom = 0.05)
   fig.subplots_adjust(right = 0.95)
   
   """
   index = 1
   # plot all success
   ax = fig.add_subplot(n_row, n_col, index)
   plot_success_scatter(ax, pid)
   index+= n_col
   # plot all times
   ax = fig.add_subplot(n_row, n_col, index)
   plot_all_finish(ax, pid)
   index+= n_col

   ax = fig.add_subplot(n_row, n_col, index)
   plot_all_postalgo(ax, pid)
   index+= n_col
   
   ax = fig.add_subplot(n_row, n_col, index)
   index+= n_col

   ax = fig.add_subplot(n_row, n_col, index)
   index+= n_col

   ax = fig.add_subplot(n_row, n_col, index)
   index+= n_col
   """
   
   plotted_cluster = list(plotted_cluster)
   LOG("Plotting cluster %s" % plotted_cluster)
   
   # plot top clusters
   for x in range(0,n_top):
      index = x + 1
      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      plot_cluster_scatter(ax, plotted_cluster[x], pid) 
   
      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      plot_cluster_finish(ax, plotted_cluster[x], pid)
      
      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      if pname != "vm_res":
         plot_cluster_ncc(ax, plotted_cluster[x], pid)

      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      if pname != "vm_res":
         plot_cluster_cctime(ax, plotted_cluster[x], pid)
      
      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      plot_cluster_postalgo(ax, plotted_cluster[x], pid)

      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      plot_cluster_allocator(ax, plotted_cluster[x], pid)

      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      if pname == "async":
         plot_cluster_allocator_wl(ax, plotted_cluster[x], pid)
      
      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      if pname != "vm_res":
         plot_cluster_ccperc(ax, plotted_cluster[x], pid)

      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      if pname != "vm_res":
         plot_cluster_ccbreakdown(ax, plotted_cluster[x], pid)

      ax = fig.add_subplot(n_row,n_col, index)
      index+= n_col
      if pname == "async":
         plot_cluster_lock(ax, plotted_cluster[x], pid)

   fig.subplots_adjust(hspace = 0.4) 
   fig.savefig(cfg.FN_PLOT + "-" + all_profiles[pid]["name"] + "-" \
      + make_time(earliest_time, date=True).replace(":","-") + ".png", format = "png")
   return

### command handler for stats
def command_handler_stats():
   LOGV("command_handler_stats")
   if not os.path.exists(cfg.FN_LOG_ALLDRS):
      command_handler_init()
      if not generate_alldrs(): return 
      if not generate_keydrs(): return
   if not os.path.exists(cfg.FN_LOG_KEYDRS):
      if not generate_keydrs(): return
   analyze()
   return

### command handler for alldrs
def command_handler_alldrs():
   LOGV("command_handler_alldrs")
   command_handler_init()
   if not generate_alldrs(): return 
   if not generate_keydrs(): return
   return   

### command handler for clean
def command_handler_clean():
   LOGV("command_handler_clean")
   if os.path.exists(cfg.DIR_WORKING):
      shutil.rmtree(cfg.DIR_WORKING)
   LOG("directories cleaned successfully")
   return

### command handler for init
def command_handler_init():
   LOGV("command_handler_init")
   os.path.exists(cfg.DIR_WORKING) or os.mkdir(cfg.DIR_WORKING)
   os.path.exists(cfg.DIR_LOG) or os.mkdir(cfg.DIR_LOG)
   for f,l in cfg.PATTERN_FILES.iteritems():
      write_list_to_file(f,l)
   LOG("drslog initialized successfully")
   return

### dispatch the command using command name
def dispatch_cmd(cmd):
   LOGV("starting...")
   LOGV("options: %s" % options)
   if cmd in ALL_COMMANDS:
      func_name = COMMAND_HANDLER_NAME_PREFIX + cmd
      func_call = None
      try:
         func_call = getattr(__import__(__name__), func_name)
      except AttributeError:
         LOGERR("command handler not found")
         PANIC()
      func_call()
      LOGV("quiting...")
   else:
      LOGERR("invalid command %s" % cmd)

### main function
def main():
   usage = "usage: %prog [options] arg"
   parser = OptionParser(usage)
   parser.add_option("-c", "--cluster", dest="cname",
                     help="the name of the cluster to be processed")
   parser.add_option("-v", "--verbose",
                     action="store_true", dest="verbose")
   parser.add_option("-q", "--quiet",
                     action="store_false", dest="verbose")
   parser.add_option("-o", "--output",
                     action="store_true", dest="output",
                     help="produce output file")
   parser.add_option("-D", "--drmdump",
                     action="store_true", dest="drmdump",
                     help="read from drmdump file")
   parser.add_option("-p", "--plot",
                     action="store_true", dest="plot",
                     help="plot the invocations")
   parser.add_option("-V", "--version",
                     dest="version", help="version of VC", default="40")
   parser.add_option("-I", "--inputdir",
                     dest="inputdir", help="input directory containing logs")
   parser.add_option("-O", "--outputdir",
                     dest="outputdir", help="output directory")
                     
   global options
   (options, args) = parser.parse_args()
   init_config()
   init_stats()
   init_patterns()
   if len(args) != 1:
       parser.error("incorrect number of arguments")
   
   cmd = args[0]
   dispatch_cmd(cmd)
   
if __name__ == "__main__":
    main()


