import sys, optparse, os, shutil
import drmplot, drmparse
import matplotlib
from matplotlib.backends.backend_agg import *
        
class TableData:
    def GetValue(self, row, col):
        rnode = self.nlist[row]
        coffset = 1
        if col == 0:
            return rnode.name
        if isinstance(rnode, drmparse.Node):
            if col == 1:
                return rnode.parent
            elif col == 2:
                return rnode.rsc['cpu'].ashares
            coffset = 3
        mystats = rnode.stats[self.stattype]
        thisvaltype = self.valtypes[col - coffset]
        thisstat = mystats[self.passnum-1][thisvaltype]
        return str(thisstat)
    
    def GetNumberRows(self):
        return len(self.nlist)
    def GetNumberCols(self):
        return len(self.valtypes) + self.coffset
    def GetRowLabelValue(self, row):
        return ""
    
    def SortData(self, keynum):
        sortkey = self.GetColLabelValue(keynum)
        if self.curkey == sortkey:
            self.sortUp = -1 * self.sortUp
        def cmpfnc(a,b):
            if sortkey == 'Name':
                akey = a.name
                bkey = b.name
            elif sortkey == 'Parent':
                akey = a.parent
                bkey = b.parent
            elif sortkey == 'CPUShares':
                akey = a.rsc['cpu'].ashares
                bkey = b.rsc['cpu'].ashares
            else:
                akey = a.stats[self.stattype][self.passnum-1][sortkey]
                bkey = b.stats[self.stattype][self.passnum-1][sortkey]
            if akey < bkey:
                return -1 * self.sortUp
            elif akey > bkey:
                return 1 * self.sortUp
            else:
                return 0
        self.curkey = keynum
        self.nlist.sort(cmpfnc)

    def GetColLabelValue(self, col):
        if col == 0:
            return "Name"
        elif self.isnode and col == 1:
            return "Parent"
        elif self.isnode and col == 2:
            return "CPUShares"
        else:
            return self.valtypes[col - self.coffset]

    def ToHtml(self):
        mystr = '<table border="box" >'
        nrows = self.GetNumberRows()
        ncols = self.GetNumberCols()
        mystr += '<tr>'
        for c in range(ncols):
            mystr += '<th>' + self.GetColLabelValue(c) + '</th>'
        mystr += '</th>'
        for r in range(nrows):
            mystr += '<tr>'
            for c in range(ncols):
                mystr += '<td>' + str(self.GetValue(r, c)) + '</td>'
            mystr += '</tr>\n'
        mystr += '</table>\n'
        return mystr

    def SetData(self, nodes, passnum):
        self.nlist = filter(lambda n: len(n.stats.keys()) > 0,
                            nodes.values())
        self.passnum = passnum
        firstnode = self.nlist[0]
        mystats = firstnode.stats[self.stattype]
        self.isnode = isinstance(firstnode, drmparse.Node)
        self.valtypes = mystats[0].keys()
        if self.isnode:
            self.coffset = 3
        else:
            self.coffset = 1
        if self.curkey:
            self.SortData(self.curkey)
            
            
    def __init__(self, nodes, passnum, stattype):
        # wxPyGridTableBase.__init__(self)
        # only consider nodes with at least some stats
        self.stattype = stattype        
        self.curkey = None
        self.sortUp = 1
        self.SetData(nodes, passnum)


class MoveTableData(TableData):
    def __init__(self, moves, passnum, maxPassMoves):
        self.maxPassMoves = maxPassMoves
        self.SetData(moves, passnum)
        self.colInfo = [
            ("VM", lambda m: m.vm.name),
            ("src", lambda m: m.src),
            ("dest", lambda m: m.dest),
            ("avgGood", lambda m: m.avgGood),
            ("peakGood", lambda m: m.peakGood),
            ("affin", lambda m: m.affin)]
        
    def SetData(self, moves, passnum):
        self.moves = moves
        self.passnum = passnum
    def GetNumberRows(self):
        return self.maxPassMoves
    def GetNumberCols(self):
        return len(self.colInfo)
    def GetValue(self, row, col):
        if row >= len(self.moves):
            return ""
        m = self.moves[row]
        colFunc = self.colInfo[col][1]
        return colFunc(m)
    
    def GetColLabelValue(self, col):
        return self.colInfo[col][0]
    def SortData(self, key):
        pass    

def MakeCanvasAgg(figure):
    return FigureCanvasAgg(figure)


class PlotManager:
    def __init__(self, parent, parser):
        self.parser = parser
        self.parent = parent

    def HandleClick(self, event):
        self.plotter.HandleClick(event)
        
    def setPlotElements(self, pelems):
        self.plotElements = pelems
        
    def doPlot(self, wantLegend):
        self.plotter = drmplot.Plotter(self.plotElements, self.parser)
        self.plotter.plotall(wantLegend=wantLegend)
        for m in self.parser.GetMoves():
            self.plotter.plotmig(m)
        self.canvas = None

    def setCanvas(self, c):
        self.canvas = c
    def getCanvas(self):
        if not self.canvas:
            self.canvas = MakeCanvasAgg(self.plotter)
        return self.canvas


class DataMgr:
    def __init__(self, inputdir):
        self.inputdir = inputdir
        self.logFname = os.path.join(inputdir, 'drmoutput.txt')
        self.resFname = os.path.join(inputdir, 'resources.txt')
        self.parser = drmparse.Parser(self.logFname, self.resFname)
        self.parser.LoadDrmFile()
        self.tables = {}        
        self.SetupTables(1, True)
        self.plotmgr = PlotManager(self, self.parser)
        self.SetupDummyConfig()

    def SetupTables(self, passnum, firstTime=False):
        stattypes = self.getParser().GetStatTypes()

        for thisstat in stattypes:
            vmNodes = self.getParser().GetNodes("vms")
            vmKey = "vms:"+thisstat
            hostNodes = self.getParser().GetNodes("hosts")
            hostKey = "hosts:"+thisstat
            if firstTime:
                self.tables[vmKey] = TableData(vmNodes,passnum,thisstat)
                self.tables[hostKey] = TableData(hostNodes,passnum,thisstat)
            else:
                self.tables[vmKey].SetData(vmNodes, passnum)
                self.tables[hostKey].SetData(hostNodes, passnum)
                
        passmoves = self.GetPassMoves(passnum)
        if firstTime:
            mpmoves = self.GetMaxPassMoves()
            self.tables["migrations"] = MoveTableData(passmoves, passnum, mpmoves)
        else:
            self.tables["migrations"].SetData(passmoves, passnum)

    def GetPassMoves(self, passnum):
        return filter(lambda m: m.passnum == passnum, self.parser.GetMoves())
    
    def GetMaxPassMoves(self):
        maxpass = self.parser.GetMaxPassnum()
        moveLens = [len(self.GetPassMoves(i)) for i in range(1,maxpass)]
        return max(moveLens)

    def doSave(self, outfile):
        outdir = os.path.dirname(outfile)
        graphicname = os.path.join(outdir,'hostgraph.png')
        self.plotmgr.doPlot(True)
        shutil.copy(self.logFname, os.path.join(outdir, 'drmlog.txt'))
        
        print 'saving to ', graphicname
        self.plotmgr.getCanvas().print_figure(graphicname, 100)

        # preserve our initial stats
        origtables = self.tables
        
        f = open(outfile, 'w')
        f.write('''
        <html>
        <head><title>DRM Test results</title></head>
        <body bgcolor="white"><a href="#tables">(skip charts)</a>
        <a href="drmlog.txt">[full log]</a><br>
        ''')
        for i in range(self.parser.GetMaxPassnum()):
            f.write('<a href="#pass%d">%d</a>,' % (i, i))
        f.write('<br>\n')

        optfile = open(os.path.join(self.inputdir, 'drmoptions.txt'), 'r')
        f.write('<b>DRM options:</b><br>\n')
        for l in optfile.readlines():
            f.write(l+'<br>')
        optfile.close()
        f.write('<br><hr><br>')
        f.write('<img src="hostgraph.png"><br>')
        allvms = self.parser.GetNodes('vms').values()
        def cmpfunc(a,b):
            if a.name > b.name:
                return 1
            elif b.name > a.name:
                return -1
            else:
                return 0
        allvms.sort(cmpfunc)
        for i in range(len(allvms) / 4):
            tmpelements = []
            thisbase = i * 4
            thisend = thisbase+4
            if thisend > len(allvms):
                thisbase = -1
            vms = allvms[4*i:min(4*(i+1),len(allvms))]
            for v in vms:
                pelem = drmplot.PlotElement(v, 'ActAvg1Min', ['run', 'cEntitled', 'cbase', 'active'])
                tmpelements.append(pelem)
            tmpplotter = drmplot.Plotter(tmpelements, self.parser)
            tmpcanvas = MakeCanvasAgg(tmpplotter)
            tmpplotter.plotall(wantLegend=True)
            # plot the initial placements
            for v in vms:
                initialHost = vstats = v.stats['ActAvg1Min'][1]['host']
                m = drmparse.Move(v, initialHost, initialHost, 0)
                tmpplotter.plotmig(m)
            for m in self.parser.GetMoves():
                tmpplotter.plotmig(m)
            outpath = os.path.join(outdir,'vmgraph%d.png' % (i))
            print 'saving to ', outpath
            tmpcanvas.print_figure(outpath, 100)
            f.write('<br><img src="vmgraph%d.png">' % (i))
        f.write('<br><hr><a name="tables"></a>')
        # print every table from every pass
        for passnum in range(2,self.parser.GetMaxPassnum()):
            f.write('<p><b>Pass %d Utilization:</b> %3.3f%%\n' % \
                    (passnum, self.getUtilization(passnum, 'ActAvg1Min')))
            self.SetupTables(passnum)
            for k in self.tables.keys():
                t = self.tables[k]
                f.write('<p><b><a name="pass%d">%s pass %d</a></b>\n' % (passnum, k, passnum))
                thtml = t.ToHtml()
                if thtml is not None:
                    f.write(t.ToHtml())
            f.write('<hr>')
                
        # return everybody to their original passnum
        self.tables = origtables
        
        f.write('</body></html>')
        f.close()

    def getParser(self):
        return self.parser
        
    def getUtilization(self, passnum, stat):
        hnodes = self.parser.GetNodes("hosts").values()
        vnodes = self.parser.GetNodes("vms").values()
        totcapacity = 0.0
        totrun = 0.0
        for h in hnodes:
            mystats = h.stats[stat][passnum]
            mycapacity = mystats["capacity"]
            totcapacity += mycapacity
        for v in vnodes:
            mystats = v.stats[stat][passnum]
            myrun = mystats["run"]
            totrun += myrun
        return 100.0 * totrun / totcapacity

    
    def SetupDummyConfig(self):
        elements = []
        for h in self.parser.hlist:
            v = self.parser.GetNode("host", h)
            pelem = drmplot.PlotElement(v, 'ActAvg1Min', ['cpuNormEnt', 'memNormEnt'])
            elements.append(pelem)
        self.plotmgr.setPlotElements(elements)

def SaveReport(inputdir, outputdir):
    matplotlib.use('Agg')
    data = DataMgr(inputdir)
    print "outdir = ", outputdir
    if not os.path.exists(outputdir):
        if not os.mkdir(outputdir):
            print "could not create directory ", outputdir
            sys.exit(-1)
    data.doSave(os.path.join(outputdir, "index.html"))
        
if __name__ == "__main__":
    opts = optparse.OptionParser()
    opts.add_option("-o", "--output", dest="outdir", default=None)
    options, args = opts.parse_args()

    inputdir = args[0]
    
    if options.outdir:
        SaveReport(inputdir, options.outdir)
    else:
        print "Need outdir!"
        sys.exit(-1)
