#!/usr/local/bin/python
#
# Runs multiple drmtest instances and summarizes the results.
#

import os.path, re, logging, time, math, sys, shutil, zipfile
import drmtest

suitelog = logging.getLogger("drmsuite")

class TestResult:
    def RecordHostState(self, hname, hstate, numvmStr):
        # replaces any previous records for this host
        # so we only keep the last values in the logfile
        numv = int(numvmStr)
        if hstate == "goingdown" and numv == 0:
            hstate = "down"
        elif hstate == "goingdown":
            hstate = "evacfail"
        self.hstates[hname] = hstate

    def RecordHostFairness(self, hname, cpufair, memfair):
        # replace previous records, as above
        self.hmemFair[hname] = float(memfair)
        self.hcpuFair[hname] = float(cpufair)

    def GetHostsDown(self):
        """Returns the number of hosts that were sucessfully evacuated"""
        return len(filter(lambda v: v=="down", self.hstates.values()))

    def GetHostEvacsFailed(self):
        """Returnst he number of hosts that we failed to evacuate"""
        return len(filter(lambda v: v=="evacfail", self.hstates.values()))

    #Function: regxMatch
    # IN: expr : the expression to be searched in the output file drmoutput.txt
    # IN: passExpr : Optional expression for pass parameter. If empty then the
    #                expr is searched in all the passes, else only at the pass
    #                which satisfies passExpr
    # OUT: Number of times a match for the expression is found
    #
    def RegxMatch(self, expr, passExpr=''):
        if passExpr == '':
           passFlag = 0
        else:
           passFlag = 1
           passExprStr = passExpr.replace('Pass', 'curPass')
           passExprStr = passExprStr.replace('pass', 'curPass')
           if (passExprStr.find('=') > -1) and (passExprStr.find('==') == -1)\
            and (passExprStr.find('>=') == -1) and (passExprStr.find('<=') == -1)\
            and (passExprStr.find('!=') == -1):
               passExprStr = passExprStr.replace('=','==')

        numrounds = int(self.drmTestResults.numrounds)
        fname = os.path.join(self.outdir, "drmoutput.txt")
        f = open(fname, "r")
        exprRe = re.compile(expr)
        passRe = re.compile("\*\*\*\*\*\*\*\*\*\*\*\*\*\*\* pass (\d+) \(time (\d+)\)"\
                            " \*\*\*\*\*\*\*\*\*\*\*\*\*\*\*")
        count = 0
        process = 0

        for line in f.xreadlines():
           if (passFlag > 0):
              m = passRe.search(line)
              if m:
                 curPass = int(m.group(1))
                 if (process == 0 and eval(passExprStr)):
                    process = 1
                    continue
                 if (process == 1 and not(eval(passExprStr))):
                    process = 0
                    continue
              if process == 0:
                 continue
           m = exprRe.search(line)
           if m:
              count += 1
        return count

    #Function: regxExtract
    # IN: expr : the expression to be searched in the output file drmoutput.txt
    # IN: passExpr : Optional expression for pass parameter. If empty then the
    #                expr is searched in all the passes, else only at the pass
    #                which satisfies passExpr
    # OUT: A list of lists. The fist list contains 1 list for each line matched
    #      and the individual lists contain the elements extracted in order depending
    #      on the input regular expression
    #
    def RegxExtract(self, expr, passExpr=''):
        if passExpr == '':
           passFlag = 0
        else:
           passFlag = 1
           passExprStr = passExpr.replace('Pass', 'curPass')
           passExprStr = passExprStr.replace('pass', 'curPass')
           if (passExprStr.find('=') > -1) and (passExprStr.find('==') == -1)\
            and (passExprStr.find('>=') == -1) and (passExprStr.find('<=') == -1)\
            and (passExprStr.find('!=') == -1):
               passExprStr = passExprStr.replace('=','==')

        fname = os.path.join(self.outdir, "drmoutput.txt")
        f = open(fname, "r")
        exprRe = re.compile(expr)
        passRe = re.compile("\*\*\*\*\*\*\*\*\*\*\*\*\*\*\* pass (\d+) \(time (\d+)\)"\
                            " \*\*\*\*\*\*\*\*\*\*\*\*\*\*\*")
        count = 0
        process = 0
        matchedElements = []
        numrounds = int(self.drmTestResults.numrounds)
        for line in f.xreadlines():
           if (passFlag > 0):
              m = passRe.search(line)
              if m:
                 curPass = int(m.group(1))
                 if (process == 0 and eval(passExprStr)):
                    process = 1
                    continue
                 if (process == 1 and not(eval(passExprStr))):
                    process = 0
                    continue
              if process == 0:
                 continue
           match = []
           match =  exprRe.findall(line)
           if match:
              matchedElements = matchedElements + match
        return matchedElements

    #Function: CompareLoadStats
    # IN: loadStat : Previous load or fairness
    # IN: load: Current load or fairness
    # OUT: 1 if the current load is equal to or within 5%
    #      of the previous load. 2 if better by >5%
    #      0 otherwise.
    #
    def CompareLoadStats(self, loadStat, load):
       if loadStat>0:
          #improvement by less than 5% is considered as same result
          if ((abs(loadStat - load))/loadStat) <= 0.05:
             return 1
          elif load < loadStat:
             return 2
          else:
             return 0
       else:
          if (load == loadStat):
             return 1
          else:
             return 0

    #Function: getGroup
    # IN: group Name
    # OUT: object which has a dictionary of all the parameters.
    # just use the using the same name as in template.ini file
    #
    def GetGroup(self, groupName):
         for group in self.drmTestResults.testgroups:
            if group['name'] == groupName:
               return group
         raise "Error in tst file. Could not Find the group"

    def compareStatsInt(self, statsLine, stats, statsName, val, statsReasonCode, statsNum, fwd):
       if statsLine.find(statsName) > -1:
          try:
             self.StorageHist[statsName] = float(stats[statsNum])
             if self.StorageHist[statsName] >= 0:
                if (fwd and (val > self.StorageHist[statsName])) or \
                        (not fwd and (self.StorageHist[statsName] < val)):
                   self.resultReason = statsReasonCode
                   self.regressed = 1
                elif (fwd and (val < self.StorageHist[statsName])) or \
                        (not fwd and (self.StorageHist[statsName] > val)):
                   self.improved = 1
             else:
                suitelog.debug("Stats does not have %s to compare"%(statsName))
                self.failStats = 1
          except:
             suitelog.debug("Stats does not have %s to compare"%(statsName))

    #Function: CompareStats
    # IN: statsName. This string is not parsed exactly and keywords
    #     cpu, mem, numMigs and paload are searched for in the string.
    # OUT: Returns 0 if the current values are better than or equal to
    #      the previous values and 1 if there is any regression
    # This function simply returns unless comparestats option is
    # set to 1
    #
    def CompareStats(self, statsName):
        improved = 0
        regressed = 0
        if not self.opts.comparestats:
           return 0
        else:
           cpuBetter = memBetter = 0
           migBetter = payloadBetter = payloadMBetter = 0
           powerBetter = fractionImbBetter = 0
           if os.path.exists(self.statfile):
              f = open(self.statfile, "r")
              statsLine = ''
              for line in f.xreadlines():
                 if line[0] == '#' or line[0] == '\n':
                    continue
                 statsLine = line
              stats = statsLine.split(',')

              if self.avgLoad["CPU"] >= 0.0:
	         cpuLoad = self.avgLoad["CPU"]
	         cpuFair = self.avgFair["CPU"]
	         memLoad = self.avgLoad["MEM"]
	         memFair = self.avgFair["MEM"]
	      else:
                 cpuLoad = self.avgCpuFair
	         cpuFair = self.varCpuFair
	         memLoad = self.avgMemFair
	         memFair = self.varMemFair

              if statsName.find('all') > -1:
                 statsName = 'cpu mem numMig payload payloadMem fractionImb'

              if statsName.find('cpu') > -1:
                 self.histCpuLoad = float(stats[1])
                 self.histCpuFair = float(stats[2])
                 cpuCompare = self.CompareLoadStats(self.histCpuLoad, cpuLoad)
                 if (self.histCpuFair >= 0) and (cpuCompare == 1):
                     cpuCompare = self.CompareLoadStats(self.histCpuFair, cpuFair)
                     if (cpuCompare == 0):
                        suitelog.debug("Test %s - CPU Fairness Stats regressed -"\
                        "Check the logfile %s for details"\
                        %(self.tname, self.testfile+'.stat'))
                        self.resultReason = "CPU"
                        cpuBetter = -1
                     elif cpuCompare == 2:
                        cpuBetter = 1
                 else:
                       self.failStats = 1


              if statsName.find('mem') > -1:
                 self.histMemLoad = float(stats[3])
                 self.histMemFair = float(stats[4])
                 memCompare = self.CompareLoadStats(self.histMemLoad, memLoad)
                 if (self.histMemFair >= 0) and (memCompare == 1):
                    memConpare =  self.CompareLoadStats(self.histMemFair, memFair)
                    if (memCompare == 0):
                       suitelog.debug("Test %s - MEM Fairness Stats regressed -"\
                       "Check the logfile %s for details"\
                       %(self.tname, self.testfile+'.stat'))
                       self.resultReason = "MEM"
                       memBetter = -1
                    elif memCompare == 2:
                       memBetter = 1
                 else:
                       self.failStats = 1


              if statsName.find('numMig') > -1:
                 self.histNumMig = int(stats[0])
                 if self.histNumMig > -1:
                    if self.numMigs > self.histNumMig:
                       suitelog.debug("Test %s - Migration Stats regressed -"\
                       "Check the logfile %s for details"\
                       %(self.tname, self.testfile+'.stat'))
                       self.resultReason = "MIG"
                       migBetter = -1
                    elif self.numMigs < self.histNumMig:
                       migBetter = 1
                 else:
                       self.failStats = 1


              if statsName.find('payload') > -1:
                 self.histPayload = float(stats[6])
                 if self.histPayload >= 0:
                    if self.histPayload > self.totRscUsage:
                       suitelog.debug("Test %s - payload Stats regressed -"\
                       "Check the logfile %s for details"\
                       %(self.tname, self.testfile+'.stat'))
                       self.resultReason = "PAYL"
                       payloadBetter = -1
                    elif self.totRscUsage > self.histPayload:
                       payloadBetter = 1
                 else:
                       self.failStats = 1

              if statsName.find('payloadMem') > -1:
                 try:
                    self.histPayloadMem = float(stats[7])
                    if self.histPayloadMem >= 0:
                       self.memPayloadPresent = 1
                       if self.histPayloadMem > self.totRscUsageMem:
                          suitelog.debug("Test %s - Memory payload Stats regressed -"\
                          "Check the logfile %s for details"\
                          %(self.tname, self.testfile+'.stat'))
                          self.resultReason = "PAYLM"
                          payloadMBetter = -1
                       elif self.totRscUsageMem > self.histPayloadMem:
                          payloadMBetter = 1
                    else:
                       self.failStats = 1
                 except:
                    suitelog.debug("Stats does not have memory payload to compare")

              if statsName.find('power') > -1:
                 try:
                    self.histPower = float(stats[8])
                    if self.histPower >= 0:
                       self.powerStatsPresent = 1
                       if self.histPower < self.totPwrUsage:
                          suitelog.debug("Test %s - Power Stats regressed -"\
                          "Check the logfile %s for details"\
                          %(self.tname, self.testfile+'.stat'))
                          self.resultReason = "PWR"
                          powerBetter = -1
                       elif self.totPwrUsage < self.histPower:
                          powerBetter = 1
                    else:
                       self.failStats = 1
                 except:
                    suitelog.debug("Stats does not have power usage to compare")

              if statsName.find('fractionImb') > -1:
                 try:
                    self.histFractionImb = float(stats[9])

                    if (round(self.histFractionImb, 4) <
                        round(self.fractionImb, 4)):
                       suitelog.debug("Test %s - Fractional Imbalance " \
                                      "regressed - Check the logfile %s " \
                                      "for details" % \
                                      (self.tname, self.testfile + '.stat'))
                       self.resultReason = "IMB"
                       fractionImbBetter = -1
                    elif (round(self.histFractionImb, 4) >
                          round(self.fractionImb, 4)):
                       fractionImbBetter = 1
                 except:
                    suitelog.debug("Stats does not have fractional imbalance to compare")

              self.regressed = regressed
              self.improved = improved
              self.compareStatsInt(statsName, stats, 'numMigsFailed', self.numMigsFailed, "FMIG", 1, 0)
              self.compareStatsInt(statsName, stats, 'numSuspended', self.numSuspended, "SUSP", 2, 0)
              self.compareStatsInt(statsName, stats, 'storPayload', self.storPayload,"PAYL", 3, 1)
              self.compareStatsInt(statsName, stats, 'storOverhead', self.storOverhead,"OVHD", 4, 1)
              self.compareStatsInt(statsName, stats, 'storUtil', self.storUtil,"EUTIL", 5, 1)
              self.compareStatsInt(statsName, stats, 'storIops', self.storIops,"IOPS", 6, 1)
              self.compareStatsInt(statsName, stats, 'storLatency', self.storLatency,"LAT", 7, 1)
              regressed = self.regressed
              improved = self.improved

        if self.failStats:
           return 0

        if (cpuBetter == -1) or (memBetter == -1) or \
           (migBetter == -1) or (payloadBetter == -1) or \
           (payloadMBetter == -1) or (powerBetter == -1) or \
           (fractionImbBetter == -1):
           regressed = 1

        if (cpuBetter == 1) or (memBetter == 1) or \
           (migBetter == 1) or (payloadBetter == 1) or \
           (payloadMBetter == 1) or (powerBetter == 1) or \
           (fractionImbBetter == 1):
           improved = 1

        if (improved and regressed):
           self.failStats = 3
           return 0

        if (improved and not regressed):
           self.failStats = 2
           return 0

        if (not improved and regressed):
           return 1


    #Function: RecordStats
    # IN: none
    # OUT: none
    # This function simply returns unless recordstats option is
    # set greater than 0. If recordstats is 1, then the stats are
    # appended to the testfilename.stat else it its truncated if
    # recordstats = 2.
    #
    def RecordStats(self):
       if self.opts.recordstats:
          if self.opts.recordstats == 2:
             f = open(self.statfile, "w")
             if self.drmTestResults.storageSimulation:
                f.write("#numMigs, numMigsFailed, numSuspended, payload, ovhd, "
                        "endUtil, iops, latency, date_time_stamp\n")
             else:
                f.write("#numMigs, cpuLoad, cpuFair, memLoad, memFair, "
                        "avgPropTime, payload, payloadMem, power, fractionImb, "
                        "date_time_stamp\n")
          else:
             f = open(self.statfile, "a")

          if self.avgLoad["CPU"] >= 0.0:
             cpuLoad = self.avgLoad["CPU"]
             cpuFair = self.avgFair["CPU"]
             memLoad = self.avgLoad["MEM"]
             memFair = self.avgFair["MEM"]
          else:
             cpuLoad = self.avgCpuFair
             cpuFair = self.varCpuFair
             memLoad = self.avgMemFair
             memFair = self.varMemFair
          statsLine = ""
          if self.drmTestResults.storageSimulation:
             statsLine = str(self.numMigs) + ',' +       \
                         str(self.numMigsFailed) + ',' + \
                         str(self.numSuspended) + ',' +  \
                         str(self.storPayload) + ',' +   \
                         str(self.storOverhead) + ',' +  \
                         str(self.storUtil) + ',' +      \
                         str(self.storIops) + ',' +      \
                         str(self.storLatency) + ',' +   \
                         time.strftime('%d %b %Y %H:%M:%S', time.localtime()) + '\n'
          else:
             statsLine = str(self.numMigs) + ',' +        \
                         str(cpuLoad) + ',' +             \
                         str(cpuFair) + ',' +             \
                         str(memLoad) + ',' +             \
                         str(memFair) + ',' +             \
                         str(self.avgPropTime) + ',' +    \
                         str(self.totRscUsage) + ',' +    \
                         str(self.totRscUsageMem) + ',' + \
                         str(self.totPwrUsage) + ',' +    \
                         str(self.fractionImb) + ',' +      \
                         time.strftime('%d %b %Y %H:%M:%S', time.localtime()) + '\n'
          f.write(statsLine)


    def ComputeSummary(self):
        """
        computes avg, var
        """
        # compute average
        tfaircpu = 0.0
	tfairmem = 0.0
        nhosts = 0
        for h in self.hcpuFair.keys():
            # only includes hosts in the ok state
            if self.hstates[h] == "ok":
                tfaircpu += self.hcpuFair[h]
		tfairmem += self.hmemFair[h]
                nhosts += 1

        if nhosts > 0:
            self.avgCpuFair = tfaircpu / nhosts
	    self.avgMemFair = tfairmem / nhosts
            # now compute sttddev
            tdevcpu = 0.0
	    tdevmem = 0.0
            for h in self.hcpuFair.keys():
                if self.hstates[h] == "ok":
                    thisdiff = self.hcpuFair[h] - self.avgCpuFair
                    tdevcpu += thisdiff * thisdiff
                    thisdiff = self.hmemFair[h] - self.avgMemFair
                    tdevmem += thisdiff * thisdiff
            self.varCpuFair = math.sqrt(tdevcpu / nhosts)
            self.varMemFair = math.sqrt(tdevmem / nhosts)
        else:
            # no valid hosts
            self.avgCpuFair = 0.0
	    self.avgMemFair = 0.0
            self.varCpuFair = 0.0
	    self.varMemFair = 0.0

        # Python output checking code for specific tests
        execCode = ''
        checkerfile = self.testfile.replace(self.tname,                     \
                                            "checkers/" + self.tname) + ".py"
        if os.path.exists(checkerfile):
           f = open(checkerfile, "r")
           for line in f.xreadlines():
              execCode += line

        # Python output checking code for all tests
        execCodeDef = ''
        df = self.testfile.replace(self.tname + ".tst", "checkers/default.py")
        if os.path.exists(df):
           f = open(df, "r")
           for line in f.xreadlines():
              execCode += line

        #Copy the class variables and functions to local so
        #that the external stub code can use it. This also
        #serves as a list of variables and function avialable
        #to the stub code
        #
        #exported functions
        RegxMatch = self.RegxMatch
        RegxExtract = self.RegxExtract
        GetGroup = self.GetGroup
        CompareStats = self.CompareStats
        RecordStats = self.RecordStats
        #exported variables
	if self.avgLoad["CPU"] >= 0.0:
	    cpuLoad = self.avgLoad["CPU"]
	    cpuFair = self.avgFair["CPU"]
	    memLoad = self.avgLoad["MEM"]
	    memFair = self.avgFair["MEM"]
	else:
            cpuLoad = self.avgCpuFair
	    cpuFair = self.varCpuFair
	    memLoad = self.avgMemFair
	    memFair = self.varMemFair

        nDown = self.GetHostsDown()
        evacTotal = nDown + self.GetHostEvacsFailed()
        succPlace = self.numSucceedPlace
        placeTotal =  self.numFailPlace + succPlace
        numMig = self.numMigs
        avgPropTime = self.avgPropTime
        payload = self.totRscUsage
        payloadMem = self.totRscUsageMem
        clusterCpuCapacity = self.clusterCpuCapacity
        numrounds = int(self.drmTestResults.numrounds)
        numMigsFailed = self.numMigsFailed
        numSuspended = self.numSuspended
        storIops = self.storIops
        storLatency = self.storLatency
        storPayload = self.storPayload
        storOverhead = self.storOverhead
        storUtil = self.storUtil
        fail = 0
        modifiers = {}
        #Execute code from tst checking files
        exec(execCode)
        exec(execCodeDef)
        #the stub code can change only the fail code
        self.fail = fail
        if fail:
           self.resultReason = "COND"
        #Check Stats
        if self.drmTestResults.storageSimulation:
           execCode = """if CompareStats("numMig, numMigsFailed, numSuspended, storIops, storLatency, storPayload, storOverhead, storUtil"):\n"""
        elif self.opts.power:
           execCode = """if CompareStats("cpu, mem, numMig, payload, payloadMem, power, fractionImb"):\n"""
        else:
           execCode = """if CompareStats("cpu, mem, numMig, payload, payloadMem"):\n"""
        if self.opts.comparestats:
           execCode += """   fail = 1\n"""
           exec(execCode)
           self.fail = fail
        #if the option is set, record stats
        if self.opts.recordstats == 2:
           RecordStats()
        if self.opts.recordstats == 1:
           #Record only if the result is not same
           #or if the previous stat does not have the newly added
           #memory payload stats
           if self.drmTestResults.storageSimulation:
              if fail > 0 or (fail == 0 and self.failStats > 0):
                 RecordStats()
           elif fail > 0 or (fail == 0 and self.failStats > 0) or (self.memPayloadPresent == 0) or\
                                   (self.powerStatsPresent == 0):
              RecordStats()

    def ParseLogs(self):
        """
        Extracts info about fairness, numMigs, etc. from drmoutput log.
        """
        fname = os.path.join(self.outdir, "drmoutput.txt")
        f = open(fname, "r")

        migRe = re.compile("moving vm (\S+) from host (\S+) to (\S+)")
        hstateRe = re.compile("host (\S+)  state:(\S+)  vms:(\d+)")
        hfairRe = re.compile("host (\S+) cpuFairness: (\S+)  memFairness: (\S+)")
        apropRe = re.compile("avgPropTime:   (\S+)")
        aloadRe = re.compile("Resource (\S+) avgLoad:   (\S+)")
        afairRe = re.compile("Resource (\S+) avgFair:   (\S+)")
        usageRe = re.compile("TotalPayload (\S+) TotalOverhead (\S+) TotalIdle (\S+) "\
                             "TotalMemoryPayload (\S+) TotalOverheadMem (\S+) "\
                             "TotalIdleMem (\S+) TotalPower (\S+)")
        migStorRe = re.compile("starting storage migration of vm")
        migStorFailRe = re.compile("FAILED storage migration of vm")
        suspendedRe = re.compile("VM (\S+) SUSPENDED due to lack of disk")
        mapRe = re.compile("MapVm (\S+) for vm (\S+)")
        mleakRe = re.compile("client block at");
        storUsageRe = re.compile("TotalPayload (\S+) TotalOverhead (\S+) UtilAtEnd (\S+) Iops (\S+) avgLatency (\S+)")
        imbalanceRe = re.compile("Current Imbalance (\S+), Target Imbalance (\S+) ")

        self.numMigs = 0
        self.numFailPlace = 0
        self.numSucceedPlace = 0
        self.numMigsFailed = 0
        self.numSuspended = 0
        self.hasLeak = False
        self.storPayload = 0.0
        self.storOverhead = 0.0
        self.storUtil = 0.0
        self.storIops = 0.0
        self.storLatency = 0.0

        # Used for running average calculations of imbalance
        imb_count = 0
        for line in f.xreadlines():
            m = migRe.search(line)
            if m:
                self.numMigs += 1
                continue
            m = hstateRe.search(line)
            if m:
                self.RecordHostState(m.group(1), m.group(2), m.group(3))
                continue
            m = hfairRe.search(line)
            if m:
                self.RecordHostFairness(m.group(1), m.group(2), m.group(3))
                continue
            m = apropRe.search(line)
            if m:
                self.avgPropTime = float(m.group(1))
		continue
            m = aloadRe.search(line)
            if m:
                self.avgLoad[m.group(1)] = float(m.group(2))
		continue
            m = afairRe.search(line)
            if m:
                self.avgFair[m.group(1)] = float(m.group(2))
		continue
            m = usageRe.search(line)
            if m:
                self.totRscUsage = float(m.group(1))
                self.totRscUsageMem = float(m.group(4))
                self.totPwrUsage = float(m.group(7))
		continue
            m = mapRe.search(line)
            if m:
                mresult = m.group(1)
                mvm = m.group(2)
                suitelog.debug("mapvm %s for vm %s" % (mresult, mvm))
                if mresult == "failed":
                    self.numFailPlace += 1
                elif mresult == "succeeded":
                    self.numSucceedPlace +=1
                else:
                    assert False, "could not understand mapvm result %s" % mresult
		continue
            m = mleakRe.search(line)
            if m:
                self.hasLeak = True
            m = migStorRe.search(line)
            if m:
                self.numMigs += 1
                continue
            m = migStorFailRe.search(line)
            if m:
                self.numMigsFailed += 1
                continue
            m = suspendedRe.search(line)
            if m:
                self.numSuspended += 1
                continue
            m = storUsageRe.search(line)
            if m:
               self.storPayload = float(m.group(1))
               self.storOverhead = float(m.group(2))
               self.storUtil = float(m.group(3))
               self.storIops = float(m.group(4))
               self.storLatency = float(m.group(5))
               continue
            m = imbalanceRe.search(line)
            if m:
               currImb = float(m.group(1))
               targetImb = float(m.group(2))
               if (targetImb == 0 or targetImb == (1 << 31) or
                   currImb == (1 << 31)):
                  continue
               else:
                  # Sample from current pass; if we do better than target,
                  # ignore the negative number.
                  fractionImb = max((currImb - targetImb) / targetImb, 0)

                  # Cumulative moving average
                  self.fractionImb = (self.fractionImb +
                                      ((fractionImb - self.fractionImb) /
                                       (imb_count + 1)))
                  imb_count += 1
               continue

    def __init__(self, outdir, tname, testfile, drmTestResults, opts):
        """Constructor that also parses logs and sets everything up"""
        self.outdir = outdir
        self.tname = tname
        self.testfile = testfile
        self.statfile = self.testfile.replace(self.tname,                  \
                                              "statfiles/" + self.tname) + \
                        ".stat.%s" % opts.ostype
        self.drmTestResults = drmTestResults
        self.opts = opts
        self.avgPropTime = -1.0
        self.avgLoad = {}
        self.avgFair = {}
        self.avgLoad["CPU"] = -1.0
        self.avgLoad["MEM"] = -1.0
        self.avgFair["CPU"] = -1.0
        self.avgFair["MEM"] = -1.0
        self.avgCpuFair = -1.0
        self.varCpuFair = -1.0
        self.avgMemFair = -1.0
        self.varMemFair = -1.0
        self.hcpuFair = {}
        self.hmemFair = {}
        self.hstates = {}
        self.fail = 0
        self.resultReason = ''
        self.failStats = 0
        self.memPayloadPresent = 0
        self.powerStatsPresent = 0
        self.clusterCpuCapacity = 0
        self.totRscUsage = 0.0
        self.totRscUsageMem = 0.0
        self.totPwrUsage = 0.0
        self.histCpuLoad = 0.0
        self.histCpuFair = 0.0
        self.histMemLoad = 0.0
        self.histMemFair = 0.0
        self.histNumMig = 0
        self.histPayload = 0.0
        self.histPayloadMem = 0.0
        self.histPower = 0.0
        self.histFractionImb = 0.0
        self.StorageHist = {}
        self.improved = 0
        self.regressed = 0
        self.fractionImb = 0.0

        for i in range(len(drmTestResults.testhosts)):
           host = drmTestResults.testhosts[i]
           self.clusterCpuCapacity += int(host['cpusize']) * int(host['numcpus'])

        self.ParseLogs()
        self.ComputeSummary()
        if self.drmTestResults.storageSimulation == 0 and self.avgPropTime == -1.0:
            assert False, "Could not compute avg proposal time"


class TestSuite:
   def __init__(self, opts, testfiles):
      self.opts = opts
      self.testfiles = testfiles
      self.foundLeaks = []

   def ShowResults(self, outdir, tname, testfile, drmTestResults, opts):
      """
      Print a line summarizing the results of the test that has run and log to
      outdir
      """
      thisRes = TestResult(outdir, tname, testfile, drmTestResults, opts)

      if thisRes.avgLoad["CPU"] >= 0.0:
         cpuLoad = thisRes.avgLoad["CPU"]
         cpuFair = thisRes.avgFair["CPU"]
         memLoad = thisRes.avgLoad["MEM"]
         memFair = thisRes.avgFair["MEM"]
      else:
         cpuLoad = thisRes.avgCpuFair
         cpuFair = thisRes.varCpuFair
         memLoad = thisRes.avgMemFair
         memFair = thisRes.varMemFair

      nDown = thisRes.GetHostsDown()
      evacTotal = nDown + thisRes.GetHostEvacsFailed()
      succPlace = thisRes.numSucceedPlace
      placeTotal = thisRes.numFailPlace + succPlace

      result = False
      if thisRes.hasLeak:
         self.foundLeaks.append(os.path.join(outdir, "drmoutput.txt"))
      if thisRes.fail == 1:
         if self.opts.comparestats:
            resultStr = thisRes.resultReason
         else:
            resultStr = "FAIL"
      else:
         if self.opts.comparestats:
            if thisRes.failStats == 1:
               resultStr = "Incomp"
            elif thisRes.failStats == 2:
               resultStr = "Better"
            elif thisRes.failStats == 3:
               resultStr = "Diff"
            else:
               resultStr = "Same"
               result = True
         else:
            resultStr = "Pass"
            result = True

      if drmTestResults.storageSimulation:
         print "  %-30s %-4d  %4d     %4d    %2.4f  %2.4f  %2.4f   %6.0f  %5.2f %6s" % \
               (tname, thisRes.numMigs, thisRes.numMigsFailed,
                thisRes.numSuspended, thisRes.storPayload, thisRes.storOverhead,
                thisRes.storUtil, thisRes.storIops, thisRes.storLatency,
                resultStr)
         if (self.opts.comparestats and (resultStr != "Same") and
             (resultStr != "COND")):
            print "  %-30s %-4d  %4d     %4d    %2.4f  %2.4f  %2.4f   %6.0f  %5.2f " \
                  "%6s" % \
                  (tname, thisRes.histNumMig,
                   thisRes.StorageHist['numMigsFailed'],
                   thisRes.StorageHist['numSuspended'],
                   thisRes.StorageHist['storPayload'],
                   thisRes.StorageHist['storOverhead'],
                   thisRes.StorageHist['storUtil'],
                   thisRes.StorageHist['storIops'],
                   thisRes.StorageHist['storLatency'], "HIST")

      elif self.opts.power:
         print "  %-30s  %-4d %6.3f   %6.3f    %2.4f   %2.4f    %7.4f  %10.6f  " \
               "%6s" % \
               (tname, thisRes.numMigs, cpuFair, memFair, thisRes.totRscUsage,
                thisRes.totRscUsageMem, thisRes.totPwrUsage,
                thisRes.fractionImb, resultStr)
         if self.opts.comparestats and (resultStr != "Same"):
            print "  %-30s  %-4d %6.3f   %6.3f    %2.4f   %2.4f    %7.4f  %10.6f  " \
                  "%6s" % \
                  (tname, thisRes.histNumMig, thisRes.histCpuFair,
                   thisRes.histMemFair, thisRes.histPayload,
                   thisRes.histPayloadMem, thisRes.histPower,
                   thisRes.histFractionImb, "HIST")
      else:
         print "  %-30s  %-4d  %6.3f %6.3f   %6.3f %6.3f   %2.4f  %2.4f  %6s" % \
               (tname, thisRes.numMigs, cpuLoad, cpuFair, memLoad, memFair,
                thisRes.totRscUsage, thisRes.totRscUsageMem, resultStr)
      return result

   def PathToTname(self, path):
      # Strip leading path and extension
      return os.path.splitext(os.path.basename(path))[0]

   def FixOptions(self):
      baseDir = self.opts.base
      if baseDir != "NOT_SET":
         opts = self.opts
         newBase = baseDir
         opts.defaultsfile = os.path.join(newBase, opts.defaultsfile)
         opts.hostsfile = os.path.join(newBase, opts.hostsfile)
         opts.vmfile = os.path.join(newBase, opts.vmfile)
         if sys.platform.find("inux") == -1:
            opts.ostype = "Win32"
         # Load the test files
         testList = self.testfiles
         doNotReplayList = []
         if len(self.testfiles) == 0:
            execCode = ""
            f = open(os.path.join(newBase,"test_list.py"),"r")
            for line in f.xreadlines():
                    execCode += line
            exec(execCode)
            testList =  DRSList
            if self.opts.storageSimulation:
               testList = SDRSList
            if self.opts.drsstress:
               testList = DRSStressList
            if self.opts.drsscale:
               testList = DRSScaleList
            doNotReplayList = DRSReplayFailed + SDRSReplayFailed
            self.opts.replay = 1
         self.testfiles = []
         self.doNotReplayList = []
         testList.sort()
         for test in testList:
            testName = test + ".tst"
            fullTestName = os.path.join(newBase, testName)
            self.testfiles.append(fullTestName)
            if test in doNotReplayList:
               self.doNotReplayList.append(fullTestName)


   def RunSuite(self):

      self.FixOptions()
      # Run all tests in self.testfiles and display results
      baseoutdir = self.opts.outdir

      # Print header
      if self.opts.storageSimulation:
         print "  %-30s nmig  failMig  vmhung  writes  ovhd    spcUtil   Iops   Lat   " \
               "  result" % \
               ("test")
      elif self.opts.power:
         print "  %-30s  nmig  cpuFair  memFair  payload  payloadM   power    " \
               "fractionImb  result" % \
               ("test")
      else:
         print "  %-30s  nmig  cpuLoad  cpuFair  memLoad  memFair  payload " \
               "payloadM result" % \
               ("test")

      for t in self.testfiles:
         for n in range(self.opts.repeat):
            # Create subdirectory for test
            tname = self.PathToTname(t)
            hosts = self.PathToTname(self.opts.hostsfile)
            defaults = self.PathToTname(self.opts.defaultsfile)
            outdir = tname

            if defaults != "default_opts":
               outdir = outdir + "-" + defaults

            if hosts != "default_hosts":
               outdir = outdir + "-" + hosts

            suitelog.info("Launching test: %s" % tname)
            newoutdir = os.path.join(baseoutdir, outdir)
            if not os.path.exists(newoutdir):
               os.mkdir(newoutdir)

            # If precomputed stats are present, put them where they will be
            # actually used.
            precomputed_stats = t + ".stats.zip"
            if os.path.exists(precomputed_stats):
               precomputed_dest = os.path.join(newoutdir, "stats.d")
               if os.path.exists(precomputed_dest):
                  shutil.rmtree(precomputed_dest)
               os.mkdir(precomputed_dest)
               f = zipfile.ZipFile(precomputed_stats)
               for info in f.infolist():
                  fname = info.filename
                  if fname.startswith('stats.d/') and fname.endswith('.txt'):
                     f.extract(info, newoutdir)
               f.close()

            self.opts.outdir = newoutdir

            # drmtest may change directory so save and restore cwd around it
            savedir = os.getcwd()
            drmTestResult = drmtest.RunDrmTest(self.opts, t)
            os.chdir(savedir)
            # Parse logs and display some data about the test
            if self.ShowResults(newoutdir, tname, t, drmTestResult, self.opts):
               os.remove(os.path.join(newoutdir, "drmoutput.txt"))
            if self.opts.replay and t not in self.doNotReplayList:
               drmReplayResult = drmtest.RunDrmReplayer(self.opts, t)

      if len(self.foundLeaks) > 0:
         print
         print "  *** ERROR You have introduced a memory leak! ***"
         print "  For more details, read the tail of one of the following " \
               "log files:"
         for l in self.foundLeaks:
            print "    ", l
         print
         print
         return -1
      else:
         return 0

if __name__ == "__main__":
    # supports all the same args as drmtest, but
    # can take multiple testfile arguments at the end
    starttime = time.time()
    opts = drmtest.GetOptionParser()
    options, testfiles = opts.parse_args()

    # log basic stuff to file only
    logging.basicConfig(level=logging.DEBUG,
                        filename=os.path.join(options.outdir, 'drmsuite.log'),
                        filemode='w')

    tSuite = TestSuite(options, testfiles)
    res = tSuite.RunSuite()

    suitelog.debug("total suite time: %.2f sec" % (time.time() - starttime))
    sys.exit(res)

