#!/usr/local/bin/python
#
# Test infrastructure for dummydrm
#

import ConfigParser, optparse, time, logging, sys, copy
import shutil, os, glob, re, random, tempfile, subprocess
import string

#
# External scripts and binaries used by drmtest
#
VMWARE_CMD = 'vmware-cmd'
KILLALL_CMD = ("killall -9 vmkload_app; ")

if os.environ.get("OS") == "Windows_NT":
   import win32api
   PLATFORM = 'win'
   SSH_CMD = 'c:\\cygwin\\bin\\ssh.exe'
   DEFAULT_DUMMY_BIN = 'dummydrm.exe'
else:
   PLATFORM = 'linux'
   SSH_CMD = '/usr/bin/ssh'
   DEFAULT_DUMMY_BIN = 'dummydrm'

#
# Default Input files (overrideable on the commandline)
#
DEFAULT_VMS = 'setup/templates.ini'
DEFAULT_HOSTS = 'setup/default_hosts.tst'
DEFAULT_INCLUDE = 'setup/default_opts.tst'
DEFAULT_RESOURCES = 'resources.txt'
DEFAULT_RULES = 'rules.txt'
DEFAULT_SHARED_BASE = '\do\Not\use'
DEFAULT_LOGLEVEL = 3
DEFAULT_SIMVMS = 'simvms.txt'
DEFAULT_SIMHOSTS = 'simhosts.txt'
DEFAULT_SIMDISKS = 'simdisks.txt'
DEFAULT_SIMNETS = 'simnets.txt'
DEFAULT_SIMDS = 'simdatastores.txt'
DEFAULT_OS = 'Linux'
DEFAULT_LOCAL_BASE = '\do\Not\use'

#
# Test parameters and output
#
DEFAULT_OUTDIR = 'drmtestfiles'
DEFAULT_TESTLEN = 60 * 6 # seconds to run whole test
DEFAULT_PASSLEN = 90

#
# VM and host configuration
#
BASE_DSK_NAME = 'smplin'
OVHD_MEM_SIZE = 70

DEFAULT_HOST_MHZ = 2000
DEFAULT_HOST_MB = 1800
DEFAULT_HOST_PCPUS = 2

HGROUP_MIN = 0
HGROUP_MAX = -1
HGROUP_MINLIMIT = -1
HGROUP_SHARES = 1000

GROUPS_PROC_NODE = '/proc/vmware/sched/groups'

#
# Local constants
#
ADDR_LOOP_DELAY = 15 # number of seconds to wait before re-checking for ip addrs
ADDR_LOOP_TIMEOUT = 60 * 10 # if it takes 10 minutes to get ip addrs, something broke
DRM_ROOT_GROUP = 'user'
DRM_PASS_SECONDS = 90

#
# Global mutable
#
sharedbase = DEFAULT_SHARED_BASE
localbase = DEFAULT_LOCAL_BASE

# These paths are relative to the shared base
WORKLOAD_OUT_DIR = 'loads'
VM_CONFIG_DIR = 'drmtest'
IPADDR_DIR = 'ipaddrs'
RESULTS_DIR = 'results'
MASTER_VM_CFG = os.path.join(VM_CONFIG_DIR, 'drmmaster', 'master.cfg')

requiredVars = [ 'description', 'workload' ]
requiredGroupVars = [ 'cpu.shares', 'cpu.min', 'cpu.max',
                      'mem.shares', 'mem.min', 'mem.max' ]
requiredHostVars = [ 'numcpus', 'cpusize', 'memsize', 'basepower',
                     'cpupower', 'boottime', 'shutdowntime']
requiredVmdkVars = ['size', 'used', 'writeworkload', 'ioworkload',
                    'percentiles', 'shares', 'reservation', 'backing']
requiredDsVars = ['size', 'slope', 'intercept', 'fudgefactor',
                  'sioclatency', 'useonlyassource', 'maxfilesize', 'dedupe',
                  'percentiopscap']
specialGroups = [ 'host', 'idle', 'system', 'user', 'kernel', 'nursery',
                  'helper', 'drivers', 'vmkapimod' ]

drmHostStateMap = {'DRM_HOST_OK' : 1,
                   'DRM_HOST_GOING_DOWN' : 2,
                   'DRM_HOST_NOT_CONNECTED' : 3,
                   'DRM_HOST_ENTER_STANDBY' : 4,
                   'DRM_HOST_STANDBY' : 5,
                   'DRM_HOST_EXIT_STANDBY' : 6}

drmVmStateMap = {'DRM_VM_POWERED_ON' : 1,
                 'DRM_VM_POWERED_OFF' : 2,
                 'DRM_VM_SUSPENDED' : 3,
                 'DRM_VM_HOT_MIGRATING' : 4,
                 'DRM_VM_POWERING_ON' : 5,
                 'DRM_VM_SPARE' : 6,
                 'DRM_VM_STATE_UNKNOWN' : 7}

testlog = logging.getLogger("drmtest")

def getConfigPath(vmnum, host, localOs = False):
    if localOs:
        basepath = os.path.join(localbase, VM_CONFIG_DIR)
        return os.path.join(basepath, 'testvm%d' % vmnum,
                            'testvm%d_%s.cfg' % (vmnum, host))
    else:
        return('%s/%s/testvm%d/testvm%d_%s.cfg' %
               (sharedbase, VM_CONFIG_DIR, vmnum, vmnum, host))

class Node:
    def __init__(self, nodedict, defaults={}):
        self.nodedict = nodedict
        # add in values from default map
        for k in defaults.keys():
            if not self.nodedict.has_key(k):
                self.nodedict[k] = defaults[k]
    def __setitem__(self, param, val):
        self.nodedict[param] = val
    def __getitem__(self, param):
        return self.nodedict[param]
    def Get(self, attrname):
        return self.nodedict[attrname]
    def Set(self, attrname, val):
        self.nodedict[attrname] = val
    def keys(self):
        return self.nodedict.keys()

class ConfigHost(Node):
    def __init__(self, optDict, type, name):
        Node.__init__(self, optDict)
        for k in type:
            if not self.nodedict.has_key(k):
               raise Exception('Invalid %s, missing %s' % (str(name),str(k)))

class ConfigVmdk(Node):
   def __init__(self, optDict):
       Node.__init__(self, optDict)
       for k in requiredVmdkVars:
          if not self.nodedict.has_key(k):
             print k
             raise 'Invalid %s, missing %s' % ("vmdk",k)

   def GetResourceLine(self, vmnum):
      n = self.nodedict
      fullname = self.nodedict['name'] + str(vmnum)
      res = "%s %s %s "%(fullname,n['size'],n['used'])
      res += n['writeworkload'].strip('"')
      res += " "
      res += n['ioworkload'].strip('"')
      res += " "
      res += n['percentiles'].strip('"')
      res += " %s"%(n['shares'])
      res += " %s"%(n['reservation'])
      return res

   def SkipLaunch(self):
        if self.nodedict.has_key('drm.placeround'):
            return True
        else:
            return False

class ConfigGroup(Node):
    def __init__(self, optDict):
        Node.__init__(self, optDict)
        for k in requiredGroupVars:
            if not self.nodedict.has_key(k):
               raise 'Invalid group, missing %s' % (k)

    def GetResourceLine(self):
        n = self.nodedict
        res = 'group %s %s %s %s %s %s %s %s ' % (n['name'], n['parent'],
                   n['cpu.shares'], n['cpu.min'], n['cpu.max'],
                   n['mem.shares'], n['mem.min'], n['mem.max'])

        cpuExpandable = False
        memExpandable = False
        if (n.has_key('cpu.expandable') and
            (n['cpu.expandable'].lower() == 'true')):
           cpuExpandable = True
        if (n.has_key('mem.expandable') and
            (n['mem.expandable'].lower() == 'true')):
           memExpandable = True

        if ((cpuExpandable and memExpandable) or
            (n.has_key('expandable') and (n['expandable'].lower() == 'true'))):
            res += '[ allExp ]'
        elif cpuExpandable:
            res += '[ cpuExp ]'
        elif memExpandable:
            res += '[ memExp ]'
        else:
            res += '[ noExp ]'
        return res

class ConfigVM(Node):
    def __init__(self, optDict, defaultVM, isDefault=False):
        Node.__init__(self, optDict, defaultVM)
        if not isDefault:
            for k in requiredVars:
                if not self.nodedict.has_key(k):
                    raise 'Invalid VM, missing %s' % (k)

    def SkipLaunch(self):
        if self.nodedict.has_key('drm.placeround'):
            return True
        else:
            return False

    def GetResourceLine(self, vmnum, othervms=None):
        n = self.nodedict
        fullname = self.nodedict['name'] + str(vmnum)
        genericCfg = getConfigPath(vmnum, '$HOST')
        res = 'vm %s %s %s %s %s ' % (fullname,
                                      n['sched.group'].rsplit("/", 1).pop(),
                                      n['sched.cpu.shares'],
                                      n['sched.cpu.min'], n['sched.cpu.max'])
        memmax = int(n['memsize']) + OVHD_MEM_SIZE
        res += ' %s %s %d [ %s ]' % (n['sched.mem.shares'],
                                     n['sched.mem.minsize'], memmax, genericCfg)
        return res

    def GetRulesLine(self, vmnum, othervms=None):
        n = self.nodedict
        res = ""
        fullname = self.nodedict['name'] + str(vmnum)
        # handle affinity/pinning
        def GetAffinLine(self, afftype, key):
            antinum = n[key]
            if antinum == "randvm":
                # randomly select a vm number for antiaffinity
                tmprange = []
                # don't include ourselves or delayed-placement vms
                for i in range(0,len(othervms)):
                    otherNum = i+1
                    if not othervms[i].SkipLaunch() and otherNum != vmnum:
                        tmprange.append(otherNum)
                antinum = random.choice(tmprange)
            else:
                antinum = int(antinum)
            othername = othervms[antinum-1].nodedict['name'] + str(antinum)
            return '%s %s %s\n' % (afftype, fullname, othername)

        if n.has_key('drm.hostaffin'):
            res += '\nhostaffin %s %s' % (fullname, n['drm.hostaffin'])
        for key in n.keys():
           if key.find('drm.antiaffin') > -1:
              res += GetAffinLine(vmnum, 'antiaffin', key)
           if key.find('drm.affin') > -1:
              res += GetAffinLine(vmnum, 'affin', key)
        if n.has_key('drm.automation') and n['drm.automation'] == 'manual':
            res += 'automation %s manual\n' % \
                   (self.nodedict['name'] + str(vmnum))
        if n.has_key('drm.placeround'):
            pround = n['drm.placeround']
            if pround.find('rand:') != -1:
                rmax = int(pround.split(':')[1])
                pval = random.randint(1,rmax)
            else:
                pval = int(pround)
            testlog.debug('delay placement of %s to %d' % (fullname, pval))
            res += '\nplace %s %d' % (fullname, pval)
        return res

class DrmTest:
    def GetHost(self, cfg, name):
        itemlist = cfg.items('host:%s' % name)
        ndict = {}
        for k,v in itemlist:
            ndict[k] = v
        h = ConfigHost(ndict, requiredHostVars, "host")
        h.name = name
        h['name'] = name
        testlog.debug('loaded host, name=%s' % h.name)
        return h

    def GetDatastore(self, cfg, name):
        itemlist = cfg.items('datastore:%s' % name)
        ndict = {}
        for k,v in itemlist:
           ndict[k] = v
        h = ConfigHost(ndict, requiredDsVars, "datastore")
        h.name = name
        h['name'] = name
        testlog.debug('loaded datastore, name=%s' % h.name)
        return h

    def GetVM(self, cfg, name):
        itemlist = cfg.items(name)
        ndict = {}
        for k,v in itemlist:
            ndict[k] = v
        v = ConfigVM(ndict, self.defaultVM, isDefault=True)
        v.name = name
        v['name'] = name
        testlog.debug('loaded VM, name=%s' % v.name)
        return v

    def GetVmdk(self, cfg, name):
        itemlist = cfg.items('vmdk:%s' %name)
        ndict = {}
        for k,v in itemlist:
            ndict[k] = v
        v = ConfigVmdk(ndict)
        v.name = name
        v['name'] = name
        testlog.debug('loaded Vmdk, name=%s' % v.name)
        return v

    def GetGroup(self, cfg, name):
        itemlist = cfg.items('group:%s' % name)
        ndict = {}
        for k,v in itemlist:
            ndict[k] = v
        g = ConfigGroup(ndict)
        g.name = name
        g['name'] = name

        # Set the root group as the default parent. We'll update copies of
        # this object later with the correct parents.
        g['parent'] = DRM_ROOT_GROUP
        testlog.debug('loaded Group, type=%s' % g.name)
        return g

    def GetPathForGroup(self, group, grouplist):
        if group == DRM_ROOT_GROUP:
           return "host/" + DRM_ROOT_GROUP
        for g in grouplist:
           if g['name'] == group:
              path = self.GetPathForGroup(g['parent'], grouplist) + "/" + group
              return path
        return ""

    def LoadNodes(self):
        nodecfg = ConfigParser.ConfigParser()
        nodecfg.read(self.vmfile)
        nodes = nodecfg.sections()
        self.defaultVM = self.GetVM(nodecfg, 'default')
        self.vmtypes['default'] = self.defaultVM

        groupre = re.compile('group:(\S+)')
	hostre = re.compile('host:(\S+)')
        vmdkre = re.compile('vmdk:(\S+)')
        dsre = re.compile('datastore:(\S+)')

        for nodename in nodes:
           m = groupre.match(nodename)
           if m:
              thisgroup = self.GetGroup(nodecfg, m.group(1))
              self.grouptypes[m.group(1)] = thisgroup
              testlog.info('GROUPLINE: ' + thisgroup.GetResourceLine())
	      continue

	   m = hostre.match(nodename)
	   if m:
              thishost = self.GetHost(nodecfg, m.group(1))
              self.hosttypes[m.group(1)] = thishost
              testlog.info('HOSTLINE: %s %d %d %d %d %d' %
                           (thishost['name'],
                            int(thishost['cpusize']),
                            int(thishost['memsize']),
                            int(thishost['numcpus']),
                            int(thishost['basepower']),
                            int(thishost['cpupower'])))
	      continue
           m = vmdkre.match(nodename)
           if m:
              thisvmdk = self.GetVmdk(nodecfg, m.group(1))
              self.vmdktypes[m.group(1)] = thisvmdk
              testlog.info('VMDKLINE: ' + thisvmdk.GetResourceLine(0))
	      continue
           m = dsre.match(nodename)
           if m:
              thisds = self.GetDatastore(nodecfg, m.group(1))
              self.dstypes[m.group(1)] = thisds
              testlog.info('DATASTORE: ' + thisds['name'])
              continue

           thisvm = self.GetVM(nodecfg, nodename)
           self.vmtypes[nodename] = thisvm
           self.addrmap[nodename] = []
           testlog.info('VMLINE: ' + thisvm.GetResourceLine(0))

    def VmFitOnHost(self, vm, hostname):
        # Overhead MB - Take overhead MB to be 75 for now
        overheadMB = 75
        host = ''
        # find the host
        for i in range(len(self.testhosts)):
           if self.testhosts[i]['name'] == hostname:
              host = self.testhosts[i]
        if host == '':
           raise 'Error loading testfile: ' + hostname + ' not found'
        if vm.nodedict.has_key('sched.mem.minsize'):
           if (int(vm['sched.mem.minsize']) + overheadMB)\
                                            > host['unReservedMem']:
              return False
        if vm.nodedict.has_key('sched.cpu.min'):
           if int(vm['sched.cpu.min']) > \
                                   host['unReservedCpu']:
              return False
        if int(vm['numvcpus']) > int(host['numcpus']):
           return False

        if vm.nodedict.has_key('sched.mem.minsize'):
           host['unReservedMem'] -= int(vm['sched.mem.minsize']) + overheadMB
        if vm.nodedict.has_key('sched.cpu.min'):
           host['unReservedCpu'] -= int(vm['sched.cpu.min'])
        return True


    def LoadTest(self, testfile=None):
        vmlinere = re.compile('\s*vm\s+(\S+)\s+(\d+)\s+(\S+)\s+(\S+)\s*(\S*)')
        grouplinere = re.compile('\s*group\s+(\S+)\s+(\S+)\s+(\S+)')
        hostlinere = re.compile('\s*host\s+(\S+)\s+(\d+)\s*(\S*)\s*(\S*)')
        configre = re.compile('\s*(\S+)\s*=\s*(\S+)')
        includere = re.compile('#include (\S+)')
        eventre = re.compile('event:\s*(\S+)\s+(\S+)\s+(\S+)\s(.*)')
        hostsetre = re.compile('hostset\s+(\S+)\s+(\S+)')
        dssetre = re.compile('datastoreSet\s+(\S+)\s+(\S+)')
        rpsetre = re.compile('VRPmembers\s+(\S+)\s+(\S+)')

        datastorere = re.compile('\s*datastore\s+(\S+)\s+(\d+)')

        if not testfile:
            tfile = self.testfile
        else:
            tfile = testfile

	testlog.debug('Loading test file %s' % (testfile))

        f = open(tfile, 'r')
        for l in f.readlines():

            m = configre.match(l)
            if m:
		testlog.debug('Config: %s = %s' % (m.group(1), m.group(2)))

                self.drmoptions[m.group(1)] = m.group(2)
                continue

            m = includere.match(l)
            if m:
                includefile = m.group(1)
                self.LoadTest(os.path.join(os.path.dirname(tfile),
                                           includefile))
                testlog.info('including file %s' % (includefile))
                continue

            m = rpsetre.match(l)
            if m:
                hostset = m.group(1)
                hostlist = m.group(2)
                hstring = 'hostset %s %s' % (hostset, hostlist)
                self.events.append(hstring)
                continue;

            m = hostsetre.match(l)
            if m:
                hostset = m.group(1)
                hostlist = m.group(2)
                hstring = 'hostset %s %s' % (hostset, hostlist)
                self.events.append(hstring)
                continue;
            m = dssetre.match(l)
            if m:
                dsset = m.group(1)
                dslist = m.group(2)
                dstring = 'datastoreSet %s %s' % (dsset, dslist)
                self.events.append(dstring)
                continue;

            m = eventre.match(l)
            if m:
                event = m.group(1)
                target = m.group(2)
                entity = m.group(3)
                arg = m.group(4)

                estring = '%s %s %s' % (event, target, entity)

                if event == 'hostStateChange':
                   estring += ' ' + str(drmHostStateMap[arg])
                #elif event == 'resPoolAdd' or event == 'resPoolEdit':
                #   groupVars = ['location'] + requiredGroupVars
                #   for p in groupVars:
                #      varre = re.compile('.*' + p + '\s*=\s*(\S+)\s.*')
                #      match = varre.match(arg)
                #      if match:
		#		         varValue = match.group(1)
                #      else:
                #         varValue = '-1'
                #      estring += ' ' + varValue
                elif event == 'vmStateChange':
                   estring += ' ' + str(drmVmStateMap[arg])
                else:
                   estring += ' ' + arg
                self.events.append(estring)
                continue

            m = vmlinere.match(l)
            if m:
               vtype = m.group(1)
               num = int(m.group(2))
               group = m.group(3)
               starthost = m.group(4)
               startds = 'randdatastore'
               if starthost.find('/') > -1:
                  splits = starthost.split('/')
                  starthost = splits[0]
                  startds = splits[1]
               compatHostsetName1 = m.group(5)
               compatHostsetName = 'allHosts'
               if (compatHostsetName1) and (compatHostsetName1 != '#') and \
                  (compatHostsetName1.find('datastore') < 0):
                  compatHostsetName = compatHostsetName1
               compatDsSet = 'allDS'
               if (compatHostsetName1.find('datastore') > -1):
                  compatDsSet = compatHostsetName1
               if not self.vmtypes.has_key(vtype):
                  raise 'Unknown vm type ' + vtype
               if (group != DRM_ROOT_GROUP and
                   not len(filter(lambda pg: pg['name'] == group,
                                  self.testgroups))):
                  raise 'invalid placement for vm(s) of type ' + vtype

               testlog.debug('using %d vms of type %s in group %s on host %s'
                             % (num, vtype, group, starthost))
               for i in range(num):
                   v = copy.deepcopy(self.vmtypes[vtype])
                   v['sched.group'] = self.GetPathForGroup(group, self.testgroups)
                   v['compatHosts'] = compatHostsetName
                   v['compatDs'] = compatDsSet
                   self.testvms.append(v)
                   if starthost == 'randhost':
                       count = 0
                       # Should be sufficient even if only
                       # 1 host is compatible
                       maxCount = 4 * len(self.testhosts)
                       while True:
                          h = random.choice(self.hosts)
                          if self.VmFitOnHost(v, h) or\
                                   v.nodedict.has_key('drm.placeround'):
                             break
                          count += 1
                          if count > maxCount:
                             raise "Cound not find a host for "+v['name']
                   else:
                       h = starthost
                       if h != 'delayplace' and \
                                    (not v.nodedict.has_key('drm.placeround')):
                          if (not self.VmFitOnHost(v, h)):
                             raise "Invalid host specified for "+v['name']
                   self.starthosts.append(h)
                   if len(self.ds) == 0:
                      continue
                   self.startds.append(startds)
               continue

            m = grouplinere.match(l)
            if m:
               grouptype = m.group(1)
               groupname = m.group(2)
               groupparent = m.group(3)
               if not self.grouptypes.has_key(grouptype):
                  raise('unknown group type ' + grouptype)
               testlog.debug('using group type %s, named %s with parent %s' %
                             (groupname, grouptype, groupparent))
               g = copy.deepcopy(self.grouptypes[grouptype])
               g['parent'] = groupparent
               g['name'] = groupname
               g['parentpath'] = self.GetPathForGroup(groupparent, self.testgroups)
               g['path'] = g['parentpath'] + "/" + groupname
               self.testgroups.append(g)
               continue

            m = hostlinere.match(l)
            if m:
               hosttype = m.group(1)
               num = int(m.group(2))
               initstate = m.group(3)
               dsset = m.group(4)

	       testlog.debug('Host: %s %d' % (hosttype, num))

               for i in range(num):
                  h = copy.deepcopy(self.hosttypes[hosttype])
                  h['name'] = h['name'] + str(len(self.hosts) + 1)
                  if initstate:
                     h['initialstate'] = drmHostStateMap[initstate]
                  else:
                     h['initialstate'] = drmHostStateMap['DRM_HOST_OK']
                  if dsset:
                     h['datastores'] = dsset
                  else:
                     h['datastores'] = 'allDS';
                  h['unReservedCpu'] = (int(h['cpusize']) * int(h['numcpus']))
                  h['unReservedMem'] = int(h['memsize'])
                  self.testhosts.append(h)
                  self.hosts.append(h['name'])

               continue
            m = datastorere.match(l)
            if m:
               hosttype = m.group(1)
               num = int(m.group(2))
               for i in range(num):
                  h = copy.deepcopy(self.dstypes[hosttype])
                  h['name'] = h['name'] + str(len(self.ds) + 1)
                  self.testds.append(h)
                  self.ds.append(h['name'])

        f.close()

    def GenConfigs(self):
        for idx in range(len(self.testvms)):
            tv = self.testvms[idx]
            vmnum = idx+1
            newCfg = os.path.join(localbase, VM_CONFIG_DIR, 'testvm%d' % vmnum)
            newCfg = os.path.join(newCfg, 'testvm%d.cfg' % vmnum)
            dskname = '%s%d.dsk' % (BASE_DSK_NAME, vmnum)
            # append our per-vm info to the master template
            shutil.copyfile(os.path.join(localbase, MASTER_VM_CFG), newCfg)
            f = open(newCfg, 'a')
            f.write('RemoteDisplay.vnc.port = %d\n' % (5900 + vmnum))
            f.write('guestinfo.name = "%s"\n' % (tv.name + str(vmnum)))
            f.write('displayname = "%s (%d)"\n' % (tv.name, vmnum))
            f.write('scsi0:1.name = "%s"\n' % (dskname))
            f.write('guestinfo.drm.sharedbase = "%s"\n' % (sharedbase))
            for k in tv.keys():
                # write everything to the config file EXCEPT
                # the required fields
                if k not in requiredVars and k.find('drm.') != 0:
                    f.write('%s = "%s"\n' % (k, tv[k]))
            f.close()
            testlog.debug('provisioned %s as type %s' % (newCfg, tv.name))

            # create one copy of the config file for each host
            for h in self.hosts:
                perHostCfgFile = getConfigPath(vmnum, h, localOs=True)
                shutil.copyfile(newCfg, perHostCfgFile)

    def RegisterConfigs(self):
        for h in self.hosts:
            arglist = [SSH_CMD, 'root@' + h, 'rm -f /etc/vmware/vm-list']
            retcode = subprocess.call(arglist)
            testlog.debug('cleared registration for host %s, retcode=%d'
                          % (h, retcode))
        for i in range(len(self.testvms)):
            vmnum = i+1
            v = self.testvms[i]
            for h in self.hosts:
                cfgPath = getConfigPath(vmnum, h)
                testlog.debug('register vm %s, cfgpath %s on host %s'
                              % (v.name, cfgPath, h))
                arglist = [SSH_CMD, 'root@' + h,
                           'echo config \\\"%s\\\" >> /etc/vmware/vm-list'
                           % cfgPath]
                retcode = subprocess.call(arglist)
                testlog.debug('pseudo-registered %s, retcode=%d'
                              % (cfgPath, retcode))

    def LaunchVMs(self):
        for i in range(len(self.testvms)):
            vmnum = i+1
            v = self.testvms[i]
            h = self.starthosts[i]
            if v.SkipLaunch():
                testlog.debug('skipping launch of %s' % (v.name))
                continue
            testlog.debug('start vm %s (%d) on host %s' % (v.name, vmnum, h))
            cfgPath = getConfigPath(vmnum, h)
            arglist = [SSH_CMD, 'root@' + h, VMWARE_CMD, cfgPath, 'start hard']
            retcode = subprocess.call(arglist)
            testlog.debug('launched %s, retcode=%d' % (cfgPath, retcode))

    def GetNumLaunchVMs(self):
        # only include vms we intend to launch
        return len(filter(lambda x: not x.SkipLaunch(), self.testvms))

    def GetNumAddrFiles(self):
        return len(glob.glob(os.path.join(localbase, IPADDR_DIR, '*.txt')))

    def LoadAddrs(self):
        # make sure the vm has finished writing this file
        time.sleep(1)
        typere = re.compile("(\S+?)\d+$")
        addrfiles = glob.glob(os.path.join(localbase, IPADDR_DIR, '*.txt'))
        numaddrs = 0
        for af in addrfiles:
            f = open(af, 'r')
            name, addr = f.readlines()
            name = name.strip()
            vmtype = typere.match(name).group(1)
            addr = addr.strip()
            f.close()
            self.addrmap[vmtype].append(addr)
            testlog.debug('ipaddrfile %s: %s -> %s' % (af, vmtype, addr))
            numaddrs += 1
        return numaddrs

    def GenWorkload(self):
        for vmnum in range(1, len(self.testvms)+1):
            v = self.testvms[vmnum-1]
            fullname = v.nodedict['name'] + str(vmnum)
            f = open(os.path.join(localbase,
                                  WORKLOAD_OUT_DIR, fullname + '.load'), 'w')
            testlog.debug('workload is %s' % (v['workload']))
            f.write('worker %s\n' % (fullname))
            wld = v['workload'].replace('"','') + '\n'
            f.write(wld)
            f.close()

            # generate files for real (non-loadgen) phases if we have
            # keys of the form "drm.workload.pre=/exit14/testprog"
            for wtype in ["pre", "post", "run"]:
                keyname = "drm.workload." + wtype
                if not v.nodedict.has_key(keyname):
                    continue
                # strip off newline and quotation marks
                data = v[keyname].strip()
                r = re.match("\"(.*)\"", data)
                data = r.group(1)
                f = open(os.path.join(localbase, WORKLOAD_OUT_DIR, fullname + '.' + wtype), 'w')
                testlog.debug('workload type %s is (%s)' % (keyname, data))
                f.write(data)
                f.close()

    def GenDRMConfig(self):
        rscoutfile = os.path.join(self.outdir, self.resourcesfile)
        f = open(rscoutfile, 'w')
        numhosts = len(self.hosts)
        for g in self.testgroups:
            thisline = g.GetResourceLine()
            f.write(thisline + '\n')
        for i in range(len(self.testvms)):
            vm = self.testvms[i]
            vmnum = i + 1
            thisline = vm.GetResourceLine(vmnum, self.testvms)
            # testlog.debug(thisline)
            f.write(thisline + '\n')
        f.close()
        testlog.info('wrote drm resource file to %s' % (rscoutfile))

        rulesoutfile = os.path.join(self.outdir, self.rulesfile)
        f = open(rulesoutfile, 'w')
        for i in range(len(self.testvms)):
            vm = self.testvms[i]
            vmnum = i + 1
            thisline = vm.GetRulesLine(vmnum, self.testvms)
            if thisline != "":
                f.write(thisline + "\n")
        for ev in self.events:
            f.write(ev + "\n")
        f.close()
        testlog.info('wrote rules file to %s' % (rulesoutfile))

        drmoptfile = os.path.join(self.outdir, 'drmoptions.txt')
        f = open(drmoptfile, 'w')
        for name in self.drmoptions.keys():
            if not (name == 'numrounds' or name == 'PollInterval'
                                        or name == 'storageSimulation'):
               f.write('%s = %s\n' % (name,self.drmoptions[name]))
        f.close()
        testlog.info('wrote drm options to %s' % (drmoptfile))

    def Shutdown(self):
        testlog.info('killing drm')
        if PLATFORM == 'win':
            PROCESS_TERMINATE = 1
            handle = win32api.OpenProcess(PROCESS_TERMINATE, False, self.drmpid)
            win32api.TerminateProcess(handle, -1)
            win32api.CloseHandle(handle)
        else:
            testlog.error('need to kill dummydrm process')

        self.drmfout.close()

        # copy out outputfile to the outdir
        testlog.info('test finished, results in %s' % (self.outdir))

    def RunDRM(self):
        os.chdir(self.outdir)
        arglist = [self.dummybin]
        testlog.info("starting drm (%s)" %
                     self.dummybin)
        self.drmfout = open(os.path.join(self.outdir, 'drmoutput.txt'), 'w')
        subp = subprocess.Popen(arglist,stdout=self.drmfout)
        self.drmpid = subp.pid

    def ReadRemoteFile(self, host, path):
        testlog.debug('retrieving %s from host %s', path, host)
        content = subprocess.Popen([SSH_CMD, 'root@' + host,
                                   'cat %s' % path],
                                   stdout=subprocess.PIPE).communicate()[0]
        return content

    def WriteRemoteFile(self, host, content, path):
        retcode = subprocess.call([SSH_CMD, 'root@' + host,
                                  'echo "%s" > %s' % (content, path)])
        return retcode

    def CleanupDrmDumps(self):
        dumpdir = self.dumpDir
        testlog.debug('removing old dump directory: %s' % dumpdir)
        shutil.rmtree(dumpdir, ignore_errors=True)

    def CleanupOldState(self, doKill=True, doDelGroups=True):
        if doKill:
            for h in self.hosts:
                testlog.debug('kill old vms on %s' % (h))
                retcode = subprocess.call([SSH_CMD, 'root@'+h, KILLALL_CMD])
                if retcode:
                    testlog.info('killall returned %d on %s' % (retcode, h))
            time.sleep(1)

        if doDelGroups:
           for h in self.hosts:
              # First construct a list of all the existing groups on this
              # host, and delete them.
              emptyLineRe = re.compile('^$')
              groupProc = self.ReadRemoteFile(h, GROUPS_PROC_NODE)
              groupsToDelete = []
              for line in groupProc.splitlines():
                 m = emptyLineRe.match(line)
                 if m:
                    break
                 m = self.groupCpuRe.match(line)
                 if m and (m.group(1) not in specialGroups):
                    group = {}
                    group['name'] = m.group(1)
                    group['parent'] = m.group(2)
                    groupsToDelete.append(group)

              # delete them back to front (== bottom up, hopefully)
              groupsToDelete.reverse()
              for g in groupsToDelete:
                  path = self.GetPathForGroup(g['name'], groupsToDelete)
                  retcode = self.WriteRemoteFile(h, 'rm ' + path, GROUPS_PROC_NODE)
                  testlog.debug("deleted existing group %s on host %s, "
                                "retcode %d" % (path, h, retcode))

        ipNames = glob.glob(os.path.join(localbase, IPADDR_DIR, '*.txt'))
        resultNames = glob.glob(os.path.join(localbase, RESULTS_DIR, '*.txt'))
        fnames = ipNames + resultNames
        for f in fnames:
            testlog.debug('unlink %s' % f)
            try:
                os.unlink(f)
            except:
                testlog.debug('could not remove %s' % f)

    def SetupOutputDir(self):
        if not os.path.exists(self.outdir):
            os.mkdir(self.outdir)
        if not os.path.exists(self.dumpDir):
           os.mkdir(self.dumpDir)
        # mandatory dummydrm.exe will be executed directly from self.dummybin
        # non-mandatory files, just for completeness
        shutil.copyfile(self.testfile,
                        os.path.join(self.outdir, 'testconfig.tst'))
        shutil.copyfile(self.vmfile,
                        os.path.join(self.outdir, 'vmfile.ini'))

        if not os.path.exists(os.path.join(self.outdir, 'tmpfiles')):
            os.mkdir(os.path.join(self.outdir, 'tmpfiles'))

    def SetupGroups(self):
        for h in self.hosts:

           for g in self.testgroups:
               retcode = \
                  self.WriteRemoteFile(h, 'mk %s %s' % (g['name'], g['parentpath']),
                                       GROUPS_PROC_NODE)
               testlog.debug("created group %s with parent %s on host %s, "
                             "retcode %d" %
                              (g['name'], g['parent'], h, retcode))

               retcode = \
                  self.WriteRemoteFile(h, 'alloc %s cpu %d %d %d %d' %
                                       (g['path'],
                                        HGROUP_MIN, HGROUP_MAX, HGROUP_SHARES,
                                        HGROUP_MINLIMIT),
                                       GROUPS_PROC_NODE)
               testlog.debug("changed cpu alloc for group %s with parent %s "
                             "on host %s, retcode %d"
                             % (g['path'], g['parent'], h, retcode))

               retcode = \
                  self.WriteRemoteFile(h, 'alloc %s mem %d %d %d %d' %
                                       (g['path'],
                                        HGROUP_MIN, HGROUP_MAX, HGROUP_SHARES,
                                        HGROUP_MINLIMIT),
                                       GROUPS_PROC_NODE)
               testlog.debug("changed mem alloc for group %s with parent %s "
                             "on host %s, retcode %d"
                             % (g['path'], g['parent'], h, retcode))

               # readback and verify that the commands actually "took"
               groupProc = \
                  self.ReadRemoteFile(h, GROUPS_PROC_NODE +
                                      ' | grep "^.*[0-9+] %s.*[0-9+] %s"' %
                                      (g['name'], g['parent']))

               # There should be two lines, one for CPU and one for memory
               groupLines = groupProc.splitlines()
               if len(groupLines) != 2:
                  raise("failed to set up group %s on host %s, "
                        "proclines:\n%s" % (g['name'], h, groupLines))

               m = self.groupCpuRe.match(groupLines[0])
               if not m or (m.group(2) != g['parent']
                            or m.group(3) != str(HGROUP_MIN)
                            or m.group(4) != str(HGROUP_MAX)
                            or m.group(5) != str(HGROUP_MINLIMIT)
                            or m.group(6) != str(HGROUP_SHARES)):

                  raise("failed to set up CPU allocation for group %s "
                        "on host %s, procline:\n%s" %
                        (g['name'], h, groupLines[0]))

               m = self.groupMemRe.match(groupLines[1])
               if not m or (m.group(2) != g['parent']
                            or m.group(3) != str(HGROUP_MIN)
                            or m.group(4) != str(HGROUP_MAX)
                            or m.group(5) != str(HGROUP_SHARES)
                            or m.group(6) != str(HGROUP_MINLIMIT)):

                  raise("failed to set up memory allocation for group "
                        "%s on host %s, procline:\n%s" %
                        (g['name'], h, groupLines[1]))

    def __init__(self, outdir, dumpDir, testfile, defaultsfile, hostsfile,
                 vmfile, resourcesfile, rulesfile, testlen, passlen, simulate,
                 dummybin, replayBin, debug, drmopts):
        self.outdir = os.path.realpath(outdir)
        if dumpDir :
           self.dumpDir = os.path.realpath(dumpDir)
        else :
           self.dumpDir = self.outdir + "/drmdump"
        self.testfile = testfile
        self.vmfile = vmfile
        self.hostsfile = hostsfile
        self.resourcesfile = resourcesfile
        self.rulesfile = rulesfile
        self.testlen = testlen
        self.passlen = passlen
        self.simulate = simulate
        self.defaultVM = {}
        self.vmtypes = {}
        self.grouptypes = {}
        self.hosttypes = {}
        self.vmdktypes = {}
        self.dstypes = {}
        self.hosts = []
        self.ds = []
        self.events = []
        self.addrmap = {}
        self.numaddrs = 0
        self.dummybin = dummybin
        self.replayBin = replayBin
        self.drmoptions = {}
        self.numrounds = 0
        self.storageSimulation = 0
        self.debug = debug
        self.groupCpuRe = re.compile("\s+\d+\s+(\S+)\s+\d+\s+(\S+)\s+\d+\s+\d+"
                                     "\s+\d+\.\d+\s+(\S+)\s+(\S+)\s+(\S+)\s+"
                                     "\S+\s+(\S+)")
        self.groupMemRe = re.compile("\s+\d+\s+(\S+)\s+\d+\s+(\S+)\s+\d+\s+\d+"
                                     "\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)")

        if drmopts:
            for o in drmopts:
                # these options will be passed through to drmoptions.txt
                oname, oval = o.split("=")
                self.drmoptions[oname] = oval

        # testvms and starthosts are parallel arrays
        self.testvms = []
        self.starthosts = []
        self.testgroups = []
        self.testhosts = []
        self.testds = []
        self.testvmdks = []
        self.startds = []

        # Preliminary steps, reading input files
        self.SetupOutputDir()
        self.LoadNodes()
        #Preprocess test file to see if we have to load
        #the default files
        f = open(self.testfile, 'r')
        includeDefaultHosts = 1
        includeDefaultOpts = 1
        excludehost = re.compile('#exclude default_hosts')
        excludeopts = re.compile('#exclude default_opts')
        for l in f.readlines():
            m = excludehost.match(l)
            if m:
               includeDefaultHosts = 0
               continue
            m = excludeopts.match(l)
            if m:
               includeDefaultOpts = 0
               continue
        if includeDefaultOpts:
           self.LoadTest(defaultsfile)
        if includeDefaultHosts:
           self.LoadTest(hostsfile)
        self.LoadTest()

    def SetupTest(self, skipRegister=False, skipGroupSetup=False):
        # Generate all relevant config files
        if not self.simulate:
            self.GenConfigs()
            if not skipGroupSetup:
                self.SetupGroups()
            if not skipRegister:
                self.RegisterConfigs()
        self.GenDRMConfig()

    def WaitForIPAddrs(self):
        starttime = time.time()
        while self.GetNumAddrFiles() < self.GetNumLaunchVMs():
            if time.time() - starttime > ADDR_LOOP_TIMEOUT:
                # abort, because we think we'll never get these guys started
                raise('TIMEOUT waiting for %d ip addrs for %d seconds'
                      % (self.GetNumLaunchVMs(), ADDR_LOOP_TIMEOUT))
                sys.exit(1)

            testlog.info("waiting for IP addresses, got %d of %d so far",
                         self.GetNumAddrFiles(), self.GetNumLaunchVMs())
            time.sleep(ADDR_LOOP_DELAY)

        naddrs = self.LoadAddrs()
        assert naddrs >= self.GetNumLaunchVMs()

    def GetDistribution(self, type, args):
        if type == 'constant':
            distParam1 = distParam2 = int(args)
        elif type == 'uniform':
            distParam1, distParam2 = map(int, args.split(','))
        elif type == 'normal':
            distParam1, distParam2 = map(int, args.split(','))
        elif type == 'precomputed':
            distParam1 = distParam2 = 0
        else:
            assert False, "don't understand distribution: " + type
        return type, distParam1, distParam2

    def GetSimWorkload(self, vmStr):
        """Converts the workload line in the VM file to a format
        that the simulator can understand"""

        wloadStr = vmStr
        wloadStr.strip()
        wloadRePeriod = re.compile('cpu (\D+)\(([\d,]*)\) (\D+)\(([\d,]*)\) '\
                        'mem (\D+)\(([\d,]*)\) (\D+)\(([\d,]*)\)( \d+)?'\
                        '( net (\D+)\(([\d,]*)\) (\D+)\(([\d,]*)\))?')
        m = wloadRePeriod.search(wloadStr)

        loadTypeCpu = m.group(1)
        loadArgsCpu = m.group(2)
        periodTypeCpu = m.group(3)
        periodArgsCpu = m.group(4)
        loadTypeMem = m.group(5)
        loadArgsMem = m.group(6)
        periodTypeMem = m.group(7)
        periodArgsMem = m.group(8)
        period = int(m.group(9) or 0)

        loadTypeNet = m.group(11) or None
        loadArgsNet = m.group(12) or None
        periodTypeNet = m.group(13) or None
        periodArgsNet = m.group(14) or None

        if loadTypeNet != None and loadArgsNet != None and \
               periodTypeNet != None and periodArgsNet != None:
           return (self.GetDistribution(loadTypeCpu, loadArgsCpu),
                self.GetDistribution(periodTypeCpu, periodArgsCpu),
                self.GetDistribution(loadTypeMem, loadArgsMem),
                self.GetDistribution(periodTypeMem, periodArgsMem), period,
                self.GetDistribution(loadTypeNet, loadArgsNet),
                self.GetDistribution(periodTypeNet, periodArgsNet))
        else:
           return (self.GetDistribution(loadTypeCpu, loadArgsCpu),
                self.GetDistribution(periodTypeCpu, periodArgsCpu),
                self.GetDistribution(loadTypeMem, loadArgsMem),
                self.GetDistribution(periodTypeMem, periodArgsMem), period,
                None, None)


    def GetSimNetwork(self, vmStr):
        wloadStr = vmStr
        wloadStr.strip()
        uplinks = "all"
        wnetLineFull = re.compile("Reservation: (\d+), Uplinks: (\S+)")
        m =  wnetLineFull.search(wloadStr)
        if m:
           reserv = int(m.group(1))
           uplinks = m.group(2)
           return (reserv, uplinks)
        wnetLine = re.compile('(\d+)')
        m = wnetLine.search(wloadStr)
        reserv = int(m.group(1))
        return (reserv, uplinks)

    def DumpSimFiles(self):
        # generate simvms.txt
        simvmfile = os.path.join(self.outdir, DEFAULT_SIMVMS)
        f = open(simvmfile, 'w')
        simnetfile = os.path.join(self.outdir, DEFAULT_SIMNETS)
        f1 = open(simnetfile, 'w')
        memZeroShare = 20
        memNonzeroShare = 10
        dumStorageFiles = 0
        for i in range(len(self.testvms)):
            vm = self.testvms[i]
            vmnum = i + 1
            fullname = vm.nodedict['name'] + str(vmnum)
            numvcpus = int(vm.nodedict['numvcpus'])
            memmb = int(vm.nodedict['memsize'])
            if vm.SkipLaunch():
                hname = 'delayplace'
            else:
                hname = self.starthosts[i]
            compatHostset = vm.nodedict['compatHosts']
            compatDsSet = vm.nodedict['compatDs']

            for key in vm.nodedict.keys():
               if key.find('workload') > -1:
                  vmStr = vm.nodedict[key]
                  (loadCpu, periodCpu, loadMem, periodMem, period, loadNet,
                   periodNet) = self.GetSimWorkload(vmStr)
                  if (loadNet == None or periodNet == None):
                     f.write('%s %d %d %s %d %d %s %d %d %d %d %s '\
                                '%d %d %s %d %d %d %s %s\n' %
                             (fullname, numvcpus, memmb, loadCpu[0],
                              loadCpu[1], loadCpu[2], periodCpu[0],
                              periodCpu[1], periodCpu[2], memZeroShare,
                              memNonzeroShare, loadMem[0], loadMem[1],
                              loadMem[2], periodMem[0], periodMem[1],
                              periodMem[2], period, hname, compatHostset))
                  else: # Used by Network-aware DRS
                     f.write('%s %d %d %s %d %d %s %d %d %d %d %s '\
                                '%d %d %s %d %d %d %s %s %s %d %d %s %d %d\n' %
                             (fullname, numvcpus, memmb, loadCpu[0],
                              loadCpu[1], loadCpu[2], periodCpu[0],
                              periodCpu[1], periodCpu[2], memZeroShare,
                              memNonzeroShare, loadMem[0], loadMem[1],
                              loadMem[2], periodMem[0], periodMem[1],
                              periodMem[2], period, hname, compatHostset,
                              loadNet[0], loadNet[1], loadNet[2],
                              periodNet[0], periodNet[1], periodNet[2]))

               # Dump the simnets.txt simultanously
               if key.find('vnic') > -1:
                  vmStr = vm.nodedict[key]
                  (resv, uplinks) = self.GetSimNetwork(vmStr)
                  # TODO: Get and fill in the teaming information
                  f1.write('%s %d %s %s\n'% (fullname, resv, key, uplinks))

            # loop through the keys to find all vmdks
            for key in vm.nodedict:
               if key.find('vmdk') > -1:
                  vmdktype = vm.nodedict[key]
                  d = copy.deepcopy(self.vmdktypes[vmdktype.strip('"')])
                  d.nodedict['vm'] = fullname
                  self.testvmdks.append(d)
                  dumStorageFiles = 1
                  hname = self.startds[i]
                  if hname == 'randdatastore':
                     ds = random.choice(self.testds)
                     hname = ds['name']
                  d.nodedict['datastore'] = hname
                  d.nodedict['compatDs'] = compatDsSet
        f.close()

        simhostsfile = os.path.join(self.outdir, DEFAULT_SIMHOSTS)
        f = open(simhostsfile, 'w')
        for i in range(len(self.testhosts)):
            host = self.testhosts[i]
            pNicsPresent = 0
            for key in host.nodedict:
               if key.find('numpnics') > -1:
                  pNicsPresent = 1
            if not pNicsPresent:
               host['numpnics'] = "0"
               host['pniccapacity'] = "0"
            f.write('%s %d %d %d %d %d %d %d %d %d %d %s\n' %
                    (host['name'], int(host['cpusize']), int(host['memsize']),
                     int(host['numcpus']), int(host['basepower']),
                     int(host['cpupower']), int(host['boottime']),
                     int(host['shutdowntime']), int(host['numpnics']),
                     int(host['pniccapacity']),int(host['initialstate']),
                     host['datastores']))

        oldhostsfile = os.path.join(self.outdir, 'hosts.txt')
        f = open(oldhostsfile, 'w')
        for i in range(len(self.testhosts)):
            host = self.testhosts[i]
            f.write('%s\n' % (host['name']))
        f.close()
        if len(self.testds) == 0:
           return

        # produce "simdisks.txt" file
        simDiskFile = os.path.join(self.outdir, DEFAULT_SIMDISKS)
        f = open(simDiskFile, 'w')
        disknum = 0
        for i in range(len(self.testvmdks)):
           disknum += 1
           disk = self.testvmdks[i]
           if disk.SkipLaunch():
                hname = 'delayplace'
           else:
                hname = disk['datastore']
           vmname = disk['vm']
           backing = disk['backing'].strip('"')
           rscLine = disk.GetResourceLine(disknum)
           dsSet =  disk['compatDs']
           f.write('%s %s %s %s %s\n' %
                   (vmname, rscLine, backing, hname, dsSet))
        f.close()

        simDsfile = os.path.join(self.outdir, DEFAULT_SIMDS)
        f = open(simDsfile, 'w')
        for i in range(len(self.testds)):
           ds = self.testds[i]
           f.write('%s %s %s ' % (ds['name'], ds['size'], ds['slope']))
           f.write('%s %s ' % (ds['intercept'], ds['fudgefactor']))
           f.write('%s %s ' % (ds['sioclatency'], ds['useonlyassource']))
           f.write('%s ' % ds['maxfilesize'])
           ddSpec = ds['dedupe'].strip('"')
           f.write('%s ' % ddSpec)
           f.write('%s\n' % ds['percentiopscap'])
        f.close()


    def LaunchTest(self):
        self.GenWorkload()
        self.LaunchVMs()
        self.WaitForIPAddrs()

        # ok, everyone has started, give them a little extra time to
        # ensure the workload client is up and running too, then launch
        time.sleep(ADDR_LOOP_DELAY)

        # now the workload is running in the vms for real, so run DRM
        self.RunDRM()
        testlog.info('sleeping for %d seconds...' % (self.testlen))
        time.sleep(self.testlen)
        self.Shutdown()

    def RunSimulator(self, numreps, seed, migthreshold, loglevel, interval):
        os.chdir(self.outdir)
        numIter = numreps
        pInterval = interval
        if self.drmoptions.has_key('numrounds'):
           numIter = self.drmoptions['numrounds']
        self.numrounds = numIter
        if self.drmoptions.has_key('PollInterval'):
           pInterval = self.drmoptions['PollInterval']
        arglist = [self.dummybin,'-r', str(seed), '-n', str(numIter),
                 '-l', str(loglevel), '-d', str(pInterval), '-t',
                   str(migthreshold), '-D', str(self.dumpDir)]
        if self.drmoptions.has_key('storageSimulation'):
           arglist.append('-o')
           self.storageSimulation = 1
        else:
           self.storageSimulation = 0
        retcode = 0
        if self.debug and sys.platform.find("inux") > -1:
           cmd = "set args -r "
           cmd1 = "%s -n %s -l %s -d %s -t %s -D %s"%(str(seed),str(numIter),
                     str(loglevel),str(pInterval),str(migthreshold),
                     self.dumpDir)
           cmd += cmd1
           if self.storageSimulation:
              cmd += " -o"
           arglist = ["gdb", self.dummybin]
           print cmd
           gGDBProcess = subprocess.Popen(arglist)
           #gGDBProcess.stdin.write(cmd)
           sys.stdout.write(cmd)
           gGDBProcess.wait()
           return
        drmfout = open(os.path.join(self.outdir, 'drmoutput.txt'), 'w')
        testlog.info('launching simulator for %d rounds' % (numreps))
        retcode = subprocess.call(arglist, stdout=drmfout)
        if retcode != 0:
            testlog.error("error running dummy, retcode = %d" % (retcode))
        else:
            testlog.info("simulator finished successfully")


    def RunReplayer(self):
        dumpDir = self.dumpDir + "/cluster"
        outputPath = os.path.join(self.outdir, 'replayoutput.txt')
        replayfout = open(outputPath, 'w')
        errPath = os.path.join(self.outdir, 'replayerr.txt')
        replayferr = open(errPath, 'w')
        success = True
        for dumpFile in os.listdir(dumpDir):
           dumpFilePath = dumpDir + "/" + dumpFile
           argListReplay = [self.replayBin, dumpFilePath]
           replayfout.write("executing: " + ' '.join(argListReplay) + '\n')
           replayfout.flush()
           retcodeReplay = subprocess.call(argListReplay, stdout=replayfout,
                                           stderr=replayferr)
           if retcodeReplay == 0:
              os.remove(dumpFilePath)
           else:
              print("Error in replaying(" + str(retcodeReplay) + "): " +
                    dumpFilePath)
              success = False
        if success:
           os.rmdir(dumpDir)
           os.remove(outputPath)
           os.remove(errPath)
        return success

def GetOptionParser():
    opts = optparse.OptionParser()
    opts.add_option("-E", "--ostype", dest="ostype", default=DEFAULT_OS)
    opts.add_option("-o", "--output", dest="outdir",
                    default=DEFAULT_OUTDIR)
    opts.add_option("-M", "--dumpDir", dest="dumpDir",
                    default=None)
    opts.add_option("-I", "--include", dest="defaultsfile",
                    default=DEFAULT_INCLUDE)
    opts.add_option("-v", "--vms", dest="vmfile",
                    default=DEFAULT_VMS)
    opts.add_option("-H", "--hosts", dest="hostsfile",
                    default=DEFAULT_HOSTS)
    opts.add_option("-R", "--resources", dest="resourcesfile",
                    default=DEFAULT_RESOURCES)
    opts.add_option("-C", "--rules", dest="rulesfile",
                    default=DEFAULT_RULES)
    opts.add_option("-t", "--testlen", dest="testlen",
                    default=DEFAULT_TESTLEN, type="int")
    opts.add_option("-p", "--passlen", dest="passlen",
                    default=DEFAULT_PASSLEN, type="int")
    opts.add_option("-Q", "--sharedBase", dest="sharedbase",
                    default=DEFAULT_SHARED_BASE)
    opts.add_option("-B", "--base", dest="base",
                    default="NOT_SET")
    opts.add_option("-L", "--localbase", dest="localbase",
                    default=DEFAULT_LOCAL_BASE)
    opts.add_option("-N", "--noregister", dest="noreg",
                    default=False,action="store_true")
    opts.add_option("-G", "--nogroupsetup", dest="nogroupsetup",
                    default=False,action="store_true")
    opts.add_option("-s", "--prepsim", dest="prepsim",
                    default=False,action="store_true")
    opts.add_option("-d", "--dummybin", dest="dummybin",
                    default=DEFAULT_DUMMY_BIN, type="string")
    opts.add_option("-l", "--loglevel", dest="loglevel",
                    default=DEFAULT_LOGLEVEL, type="int")
    opts.add_option("-S", "--runsim", dest="runsim", default=1, type="int")
    opts.add_option("-D", "--debug", dest="debug", action='store_true',default=False)
    opts.add_option("-c", "--compile", dest="compile", action='store_true',default=False)
    opts.add_option("-1", "--stress", dest="drsstress", default=0, type="int")
    opts.add_option("-2", "--scale", dest="drsscale", default=0, type="int")
    opts.add_option("-r", "--randseed", dest="randseed", default=1,
                    type="int")
    opts.add_option("-i", "--interval", dest="interval", default=300,
                    type="int")
    opts.add_option("-O", "--drmoption", dest="drmopts", action="append",
                    type="string")
    opts.add_option("-5", "--comparestats", dest="comparestats", default=1, type="int")
    opts.add_option("-a", "--record", dest="recordstats", default=0, type="int")
    opts.add_option("-m", "--migrationthreshold", dest="migrationthreshold",
                          default=3, type="int")
    opts.add_option("-g", "--storage", dest="storageSimulation",
                    default=0, type="int")
    opts.add_option("-P", "--power", dest="power",
                    default=1, type="int")
    opts.add_option("-y", "--replay", dest="replay", default=0, type="int")
    opts.add_option("--replay-bin", dest="replayBin", type="string")
    opts.add_option("-n", "--repeat", dest="repeat", default=1, type="int")

    return opts

def RunDrmTest(opts, testfile):
    global sharedbase
    global localbase

    # allow a random seed to be specified so we can repeat tests that
    # may have random host placements, start times, etc.
    seed = opts.randseed
    migthreshold = opts.migrationthreshold
    if migthreshold <= 0:
       migthreshold = 1
    #If we are going to compare and record results, use seed = 1
    if opts.comparestats > 0 or opts.recordstats > 0:
       seed = 1
    #force comparestats if recordstats is set
    if opts.recordstats == 1:
       opts.comparestats = 1
    if not seed:
       # use system time in tenths of a second
       seed = int(time.time() * 10.0)
    random.seed(seed)
    testlog.info("random seed: %d" % (seed))
    testlog.info("Migration threshold %d"%(migthreshold))
    sharedbase = opts.sharedbase
    localbase = opts.localbase
    testrun = DrmTest(opts.outdir,
                      opts.dumpDir,
                      testfile,
                      opts.defaultsfile,
                      opts.hostsfile,
                      opts.vmfile,
                      opts.resourcesfile,
                      opts.rulesfile,
                      opts.testlen,
                      opts.passlen,
                      opts.prepsim or opts.runsim,
                      opts.dummybin,
                      opts.replayBin,
                      opts.debug,
                      opts.drmopts)

    if opts.runsim > 0 or opts.prepsim:
        testrun.SetupTest()
        testrun.DumpSimFiles()
        if opts.runsim > 0:
            testrun.CleanupDrmDumps()
            testrun.RunSimulator(opts.runsim, seed, migthreshold,
                                 opts.loglevel, opts.interval)
    else:
        testrun.CleanupOldState()
        testrun.SetupTest(skipRegister=opts.noreg,
                          skipGroupSetup=opts.nogroupsetup)
        testrun.LaunchTest()

    return testrun;   #So that we can use the results and not parse the files again

def RunDrmReplayer(opts, testfile):
    testrun = DrmTest(opts.outdir,
                      opts.dumpDir,
                      testfile,
                      opts.defaultsfile,
                      opts.hostsfile,
                      opts.vmfile,
                      opts.resourcesfile,
                      opts.rulesfile,
                      opts.testlen,
                      opts.passlen,
                      opts.prepsim or opts.runsim,
                      opts.dummybin,
                      opts.replayBin,
                      opts.debug,
                      opts.drmopts)
    testrun.RunReplayer()

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s.%(msecs)d %(levelname)s %(message)s',
                        datefmt='%H:%M:%S')
    tOpts = GetOptionParser()
    tOptions, args = tOpts.parse_args()
    testfile = args[0]
    RunDrmTest(tOptions, testfile)

