import os, re, sys
from optparse import OptionParser

from parsers import getVersionParser
from tools import getTool

class DumpParserException(Exception):

    def __init__(self, fname):
        Exception.__init__(self, "not a valid dump file: %s" % (fname))

class DumpFileParser():

    def __init__(self, filepath, strict=False, verbose=False):
        self._filepath = filepath
        self._strict = strict
        self._verbose = verbose

    def accept(self, visitor):
        f = open(self._filepath, "r")
        header = f.readline()
        if not header:
            raise DumpParserException(self._filepath)

        token, version = header.split()
        if token != "VERSION":
            raise DumpParserException(self._filepath)

        options = {}
        if self._filepath.endswith("-proposeActions.dump"):
            options['actions'] = True
        elif self._filepath.endswith("-sioMapVm.dump"):
            options['sioMapVm'] = True

        if self._verbose:
            options['verbose'] = True
        p = getVersionParser(int(version), f, strict=self._strict, **options)
        visitor.visitVersion(int(version))
        p.accept(visitor)
        remaining = len(f.readlines())
        f.close()
        if remaining is not 0:
            raise Exception("trailing data (%d lines)in file: %s" % (remaining, self._filepath))


def _listDirectory(dirpath):
    files_rxp = re.compile("(\d+)-(mapVm|proposeActions).dump")

    dirList = sorted([os.path.join(dirpath, f) for f in os.listdir(dirpath)
                      if (os.path.isfile(os.path.join(dirpath, f))
                          and files_rxp.search(f))])

    for fname in dirList:
        yield fname

if __name__ == '__main__':
    root_dir = os.path.dirname(__file__) or "."
    sys.path.append(root_dir)

    parser = OptionParser()
    parser.add_option("-d", "--directory", dest="directory",
                      help="root directory for dump files", metavar="DIR", default=".")
    parser.add_option("-f", "--file", dest="file",
                      help="dump file", metavar="FILE", default=None)
    parser.add_option("-o", "--output", dest="output",
                      help="directory to put generated files", metavar="DIR", default=".")
    parser.add_option("-t", "--tool", dest="tool",
                      help="tool to run", default="dummy")
    parser.add_option("-s", "--strict", dest="strict", action="store_true",
                      help="proceeds with parsing only if we have a parser available for the exact drmdump file version")
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true", help="get more details about what the parser is doing (mostly useful for debugging)")

    (options, args) = parser.parse_args()

    if not os.path.exists(options.output):
        os.mkdir(options.output)
    t = getTool(options.tool, output=options.output)
    if options.file is not None:
        p = DumpFileParser(options.file, strict=options.strict, verbose=options.verbose)
        p.accept(t)
    else:
        for x in _listDirectory(options.directory):
            if options.verbose:
                print "> Parsing %s" % (x)
            p = DumpFileParser(x, strict=options.strict, verbose=options.verbose)
            p.accept(t)

    t.finalize()
