from helpers import TabBasedParser

class DumpSectionException(Exception):
    pass

class ArrayParser(TabBasedParser):

    TOKEN = None

    def _visitItem(self, visitor):
        pass

    def accept(self, visitor):
        token, number = self._parseLine()
        assert token == self.TOKEN

        for _ in range(int(number)):
            self._visitItem(visitor)
        visitor.visitEnd()

class StringOptionsParser(ArrayParser):

    TOKEN = "DrmRebalanceModule::_stringOpts"

    def _visitItem(self, visitor):
        token, index = self._parseLine()
        assert token == "DrmStringOption"
        token, value = self._parseLine()
        assert token == "DrmOption"
        visitor.visitOption(value)

class IntOptionsParser(ArrayParser):

    TOKEN = "DrmRebalanceModule::_intOpts"

    def _visitItem(self, visitor):
        token, index, val = self._parseLine()
        assert token == "DrmIntOption"
        token, value = self._parseLine()
        assert token == "DrmOption"
        visitor.visitOption(value)

class EnumOptionsParser(ArrayParser):

    TOKEN = "DrmRebalanceModule::_enumOpts"

    def _visitItem(self, visitor):
        token, index, val = self._parseLine()
        assert token == "DrmEnumOption"
        token, value = self._parseLine()
        assert token == "DrmOption"
        visitor.visitOption(value)

class OptionsParser(TabBasedParser):

    STRING_OPTIONS_PARSER = StringOptionsParser
    INT_OPTIONS_PARSER = IntOptionsParser
    ENUM_OPTIONS_PARSER = EnumOptionsParser

    def accept(self, visitor):
        v = visitor.visitStringOptions()
        self.STRING_OPTIONS_PARSER(self._fp).accept(v)

        v = visitor.visitIntOptions()
        self.INT_OPTIONS_PARSER(self._fp).accept(v)

        v = visitor.visitEnumOptions()
        self.ENUM_OPTIONS_PARSER(self._fp).accept(v)

        visitor.visitEnd()

class SpiParser(TabBasedParser):

    def accept(self, visitor):
        token, poll, time, clock, wrap = self._parseLine()
        assert token == "DrmSpi"

        visitor.visitPollInterval(poll)
        visitor.visitCurrentTime(time)
        visitor.visitLastClock(clock)
        visitor.visitNumWrapArounds(wrap)

        visitor.visitEnd()

class ObjectParser(TabBasedParser):

    def _acceptHeader(self, visitor):
        token, id, path = self._parseLine()
        assert token == "DrmObject"

        visitor.visitId(id)
        visitor.visitPath(path)

    def accept(self, visitor):
        self._acceptHeader(visitor)

        token, number = self._parseLine()
        assert token == "DrmObject::_events"

        for _ in range(int(number)):
            raise Exception("Not implemented: item in %s" % (token))

class HostParser(TabBasedParser):

    OBJECT_PARSER = ObjectParser

    CAPACITIES = ["cpu", "mem"]

    def _acceptHeader(self, visitor):
        tok, vmotion, state, cores, threads, mhz, boot_time, shutdown_time, auto_level, build, name = self._parseLine()
        assert tok == "DrmHost"

        visitor.visitInVmotion(vmotion)
        visitor.visitHostState(state)
        visitor.visitNumCpuCores(cores)
        visitor.visitNumCpuThreads(threads)
        visitor.visitCpuMhz(mhz)
        visitor.visitBootTime(boot_time)
        visitor.visitShutDownTime(shutdown_time)
        visitor.visitAutoLevel(auto_level)
        visitor.visitBuild(build)
        visitor.visitName(name)

    def _acceptSections(self, visitor):
        items = self._parseLine()
        assert items[0] == "DrmCapacity"
        for c, it in zip(self.CAPACITIES, items[1:]):
            reserved, total = it.split(",")
            visitor.visitCapacity(c, total, reserved)

        v = visitor.visitObject()
        self.OBJECT_PARSER(self._fp).accept(v)

    def accept(self, visitor):
        self._acceptHeader(visitor)
        self._acceptSections(visitor)
        visitor.visitEnd()

class ConfigHostsParser(ArrayParser):

    TOKEN = "DrmDomainConfig::_hosts"
    HOST_PARSER = HostParser

    def _visitItem(self, visitor):
        v = visitor.visitHost()
        self.HOST_PARSER(self._fp).accept(v)

class ResourceAllocationParser(TabBasedParser):

    def accept(self, visitor):
        token, has_reservation, reservation_value, is_expandable, exp_value, has_limit, limit_value, has_overhead, overhead_value = self._parseLine()
        assert token == "ResourceAllocationInfo"

        if has_reservation:
            visitor.visitReservation(reservation_value)
        if is_expandable:
            visitor.visitExpandableReservation(exp_value)
        if has_limit:
            visitor.visitLimit(limit_value)
        if has_overhead:
            visitor.visitOverheadLimit(overhead_value)

        token, shares, level = self._parseLine()
        assert token == "SharesInfo"

        visitor.visitShares(shares, level)

class ResourceConfigSpecParser(TabBasedParser):

    RESOURCE_ALLOCATION_PARSER = ResourceAllocationParser

    def accept(self, visitor):
        (token,) = self._parseLine()
        assert token == "ResourceConfigSpec"

        v = visitor.visitCpuAllocation()
        self.RESOURCE_ALLOCATION_PARSER(self._fp).accept(v)

        v = visitor.visitMemAllocation()
        self.RESOURCE_ALLOCATION_PARSER(self._fp).accept(v)

class AllocParser(TabBasedParser):

    RESOURCE_CONFIG_SPEC_PARSER = ResourceConfigSpecParser

    def __init__(self, fp):
        TabBasedParser.__init__(self, fp)

    def _acceptHeader(self, visitor):
        (token,) = self._parseLine()
        assert token == "DrmAlloc"

        v = visitor.visitResourceConfigSpec()
        self.RESOURCE_CONFIG_SPEC_PARSER(self._fp).accept(v)

    def accept(self, visitor):
        self._acceptHeader(visitor)
        visitor.visitEnd()

class NodeParser(TabBasedParser):

    OBJECT_PARSER = ObjectParser
    ALLOC_PARSER = AllocParser

    def _acceptHeader(self, visitor):
        token, id, name = self._parseLine()
        assert token == "DrmNode"
        visitor.visitId(id)
        visitor.visitName(name)

    def accept(self, visitor):
        self._acceptHeader(visitor)
        v = visitor.visitAlloc()
        self.ALLOC_PARSER(self._fp).accept(v)

        v = visitor.visitObject()
        self.OBJECT_PARSER(self._fp).accept(v)

        visitor.visitEnd()

class VmParser(TabBasedParser):

    NODE_PARSER = NodeParser
    ALLOC_PARSER = AllocParser

    def _acceptHeader(self, visitor):
        tok, level, dest, on, state, vcpus, mem, overhead  = self._parseLine()
        assert tok == "DrmVm"

        visitor.visitAutomationLevel(level)
        visitor.visitDestHostId(dest)
        visitor.visitRunningOn(on)
        visitor.visitVmState(state)
        visitor.visitNumVcpus(vcpus)
        visitor.visitMem(mem)
        visitor.visitOverhead(overhead)

        for it in self._parseInlineArray("DrmVm::_compatibleHosts"):
            visitor.visitCompatibleHost(it)

        for it in self._parseInlineArray("DrmVm::_incompatibleVMs"):
            visitor.visitIncompatibleVm(it)

        for it in self._parseInlineArray("DrmVm::_affineVMs"):
            visitor.visitAffineVm(it)

    def _acceptSections(self, visitor):
        v = visitor.visitAlloc()
        self.ALLOC_PARSER(self._fp).accept(v)

        token, number = self._parseLine()
        assert token == "DrmVm::_recentVmotions"
        for _ in range(int(number)):
            raise Exception("Not implemented: item in %s" % (token))

        token, number = self._parseLine()
        assert token == "DrmVm::_predictedOverheadMap"
        for _ in range(int(number)):
            x, y = self._parseLine()
            visitor.visitPredictedOverhead(x, y)

        v = visitor.visitNode()
        self.NODE_PARSER(self._fp).accept(v)


    def accept(self, visitor):
        self._acceptHeader(visitor)
        self._acceptSections(visitor)
        visitor.visitEnd()

class ConfigVmsParser(ArrayParser):

    TOKEN = "DrmDomainConfig::_vms"
    VM_PARSER = VmParser

    def _visitItem(self, visitor):
        v = visitor.visitVm()
        self.VM_PARSER(self._fp).accept(v)

class GroupParser(TabBasedParser):

    NODE_PARSER = NodeParser
    ALLOC_PARSER = AllocParser

    def accept(self, visitor):
        token, number = self._parseLine()
        assert token == "DrmGroup::_curAllocOnHost"

        for _ in range(int(number)):
            id, alloc = self._parseLine()
            visitor.visitId(id)

            self._unparseItemsLine([alloc])
            v = visitor.visitAlloc()
            self.ALLOC_PARSER(self._fp).accept(v)

        v = visitor.visitNode()
        self.NODE_PARSER(self._fp).accept(v)

        visitor.visitEnd()

class ConfigGroupsParser(ArrayParser):

    TOKEN = "DrmDomainConfig::_groups"
    GROUP_PARSER = GroupParser

    def _visitItem(self, visitor):
        v = visitor.visitGroup()
        self.GROUP_PARSER(self._fp).accept(v)

class ConfigHostsWithStatsParser(TabBasedParser):

    def accept(self, visitor):
        for it in self._parseInlineArray("DrmDomainConfig::_hostsWithStats"):
            visitor.visitHost(it)

        visitor.visitEnd()

class ConfigVmsWithStatsParser(TabBasedParser):

    def accept(self, visitor):
        for it in self._parseInlineArray("DrmDomainConfig::_vmsWithStats"):
            visitor.visitVm(it)

        visitor.visitEnd()

class ConfigVpxPath2IdHostsVmsParser(TabBasedParser):

    def accept(self, visitor):
        token, number = self._parseLine()
        assert token == "DrmDomainConfig::_vpxPath2IdHostVM"

        for i in range(int(number)):
            id, path = self._parseLine()
            visitor.visitPathId(id, path)

        visitor.visitEnd()

class RecentStatsParser(TabBasedParser):

    def _acceptStats(self, visitor, n):
        for _ in range(n):
            token, timestamp, value = self._parseLine()
            assert token == "DrmStat"

            visitor.visitTimestamp(timestamp)
            visitor.visitValue(value)

    def accept(self, visitor):
        token, size, timeout_sec, timeout, number = self._parseLine()
        assert token == "RecentStatsCache"

        visitor.visitMaxSize(size)
        visitor.visitTimeoutSec(timeout_sec)
        visitor.visitTimeout(timeout)

        self._acceptStats(visitor, int(number))
        visitor.visitEnd()

class ObjectStatsParser(TabBasedParser):

    RECENT_STATS_PARSER = RecentStatsParser

    def __init__(self, fp):
        TabBasedParser.__init__(self, fp)
        self._descriptor = []

    def _acceptRecentStats(self, visitor, n):
        assert n == len(self._descriptor)
        for id in self._descriptor:
            v = visitor.visitRecentStats(id)
            self.RECENT_STATS_PARSER(self._fp).accept(v)

    def _setStatsDescriptor(self, descriptor):
        self._descriptor = descriptor

    def accept(self, visitor):
        token, id, size, number = self._parseLine()
        assert token == "DrmObjectStats:"

        visitor.visitId(id)
        visitor.visitSize(size)

        self._acceptRecentStats(visitor, int(number))
        visitor.visitEnd()

class HostStatsParser(ArrayParser):

    TOKEN = "DrmDomainStats::_hostStats"
    OBJECT_STATS_PARSER = ObjectStatsParser

    STATS_DESCRIPTOR = ["HostCpuActAvg1MinPct", "HostCpuActAvg5MinPct",
                        "HostCpuActAvg15MinPct", "HostCpuActPeak1MinPct",
                        "HostCpuActPeak5MinPct", "HostCpuActPeak15MinPct",
                        "HostCpuRunAvg1MinPct", "HostCpuRunAvg5MinPct",
                        "HostCpuRunAvg15MinPct", "HostSamplePeriodSec"]

    def _visitItem(self, visitor):
        items = self._parseLine()
        id = items[0]
        self._unparseItemsLine(items[1:])

        visitor.visitId(id)

        v = visitor.visitObjectStats()
        p = self.OBJECT_STATS_PARSER(self._fp)
        p._setStatsDescriptor(self.STATS_DESCRIPTOR)
        p.accept(v)

class VmStatsParser(ArrayParser):

    TOKEN = "DrmDomainStats::_vmStats"
    OBJECT_STATS_PARSER = ObjectStatsParser

    STATS_DESCRIPTOR = ["VmCpuActAvg1MinPct", "VmCpuActAvg5MinPct",
                        "VmCpuActAvg15MinPct", "VmCpuActPeak1MinPct",
                        "VmCpuActPeak5MinPct", "VmCpuActPeak15MinPct",
                        "VmCpuRunAvg1MinPct", "VmCpuRunAvg5MinPct",
                        "VmCpuRunAvg15MinPct", "VmCpuRunPeak1MinPct",
                        "VmCpuRunPeak5MinPct", "VmCpuRunPeak15MinPct",
                        "VmCpuMaxLimited1MinPct", "VmCpuMaxLimited5MinPct",
                        "VmCpuMaxLimited15MinPct", "VmMemSharedZeroMB",
                        "VmMemSharedNonzeroMB", "VmMemGrantedMB",
                        "VmMemActiveMB", "VmMemOverheadMB",
                        "VmMemBalloonMB", "VmMemSwapMB",
                        "VmHostMemUsageMB", "StatSampleCount"]

    def _visitItem(self, visitor):
        items = self._parseLine()
        id = items[0]
        self._unparseItemsLine(items[1:])

        visitor.visitId(id)

        v = visitor.visitObjectStats()
        p = self.OBJECT_STATS_PARSER(self._fp)
        p._setStatsDescriptor(self.STATS_DESCRIPTOR)
        p.accept(v)

class StatsParser(TabBasedParser):

    HOST_STATS_PARSER = HostStatsParser
    VM_STATS_PARSER = VmStatsParser

    def _acceptHeader(self, visitor):
        items = self._parseLine()
        assert items[0] == "DrmDomainStats"

        for it in items[1:]:
            visitor.visitStatId(it)

    def _acceptSections(self, visitor):
        v = visitor.visitHostStats()
        self.HOST_STATS_PARSER(self._fp).accept(v)

        v = visitor.visitVmStats()
        self.VM_STATS_PARSER(self._fp).accept(v)

    def accept(self, visitor):
        self._acceptHeader(visitor)
        self._acceptSections(visitor)
        visitor.visitEnd()

class InputProposeActionsParser(TabBasedParser):

    def accept(self, visitor):
        token, alloc = self._parseLine()
        assert token == "InputProposeActions"

        visitor.visitAllocOnly(alloc)

        visitor.visitEnd()

class GraphNodeParser(TabBasedParser):

    def accept(self, visitor):
        token, pos, number = self._parseLine()
        assert token == "GraphNode"

        visitor.visitPos(pos)

        items = self._parseLine()
        assert items[0] == "GraphNode::_prereqPos"

        for i in range(int(number)):
            visitor.visitPrereqPos(items[1+i])

        visitor.visitEnd()

class ActionType(object):
    (LOWEST_ACTION_UNUSED,
     VMOTION_ACTION,
     CHANGE_RES_SETTINGS_ACTION,
     POWER_ACTION_DRM,
     INIT_PLACE_ACTION,
     HIGHEST_ACTION_UNUSED) = range(6)

class ActionBaseParser(TabBasedParser):

    def accept(self, visitor):
        token, state = self._parseLine()
        assert token == "DrmActionBase"

        visitor.visitState(state)

class ChangeResSettingsParser(TabBasedParser):

    RESOURCE_CONFIG_SPEC_PARSER = ResourceConfigSpecParser
    ACTION_BASE_PARSER = ActionBaseParser

    def accept(self, visitor):
        token, vm, target, host, type, entity_path, host_path = self._parseLine()
        assert token == "DrmActionChangeResourceSettings"

        visitor.visitIsVm(vm)
        visitor.visitTarget(target)
        visitor.visitHost(host)
        visitor.visitResSettingsType(type)
        visitor.visitEntityPath(entity_path)
        visitor.visitHostPath(host_path)

        v = visitor.visitResourceConfigSpec()
        self.RESOURCE_CONFIG_SPEC_PARSER(self._fp).accept(v)

        v = visitor.visitActionBase()
        self.ACTION_BASE_PARSER(self._fp).accept(v)

        visitor.visitEnd()

class ActionParser(TabBasedParser):

    ACTION_TYPE = ActionType
    CHANGE_RES_SETTINGS_PARSER = ChangeResSettingsParser

    def _acceptChangeRes(self, visitor):
        v = visitor.visitChangeResSettings()
        self.CHANGE_RES_SETTINGS_PARSER(self._fp).accept(v)

    def _dispatch(self, id, visitor):
        if id == self.ACTION_TYPE.CHANGE_RES_SETTINGS_ACTION:
            self._acceptChangeRes(visitor)
        else:
            raise Exception("Not implemented: type %s" % (id))

    def accept(self, visitor):
        items = self._parseLine()
        id = int(items[0])

        self._unparseItemsLine(items[1:])
        self._dispatch(id, visitor)

class RecommendationParser(TabBasedParser):

    ACTION_PARSER = ActionParser

    def _acceptHeader(self, visitor):
        token, rating, id, state, self.__number = self._parseLine()
        assert token == "DrmRecommendation"

        visitor.visitRating(rating)
        visitor.visitId(id)
        visitor.visitState(state)

    def _acceptActions(self, visitor):
        for _ in range(int(self.__number)):
            v = visitor.visitAction()
            self.ACTION_PARSER(self._fp).accept(v)

    def _acceptReason(self, visitor):
        token, code, text = self._parseLine()
        assert token == "DrmRecReason"

        visitor.visitReasonCode(code)
        visitor.visitReasonText(text)

    def _acceptPrereq(self, visitor):
        items = self._parseLine()
        assert items[0] == "DrmRecommendation::_prerequisite"

        for it in items[1:]:
            visitor.visitPrereq(it)

    def accept(self, visitor):
        self._acceptHeader(visitor)
        self._acceptPrereq(visitor)
        self._acceptReason(visitor)
        self._acceptActions(visitor)
        visitor.visitEnd()

class ResultsProposeActionsParser(TabBasedParser):

    RECOMMENDATION_PARSER = RecommendationParser

    def accept(self, visitor):
        token, status, number = self._parseLine()
        assert token == "ResultsProposeActions"

        visitor.visitStatus(status)
        for _ in range(int(number)):
            v = visitor.visitRecommendation()
            self.RECOMMENDATION_PARSER(self._fp).accept(v)

        visitor.visitEnd()

class ProposeActionsParser(TabBasedParser):

    INPUT_PROPOSE_ACTIONS_PARSER = InputProposeActionsParser
    RESULTS_PROPOSE_ACTIONS_PARSER = ResultsProposeActionsParser

    def accept(self, visitor):
        v = visitor.visitInputProposeActions()
        self.INPUT_PROPOSE_ACTIONS_PARSER(self._fp).accept(v)

        v = visitor.visitResultsProposeActions()
        self.RESULTS_PROPOSE_ACTIONS_PARSER(self._fp).accept(v)

        visitor.visitEnd()

class ConfigParser(TabBasedParser):

    CONFIG_HOSTS_PARSER = ConfigHostsParser
    CONFIG_VMS_PARSER = ConfigVmsParser
    CONFIG_GROUPS_PARSER = ConfigGroupsParser
    CONFIG_HOSTS_WITH_STATS_PARSER = ConfigHostsWithStatsParser
    CONFIG_VMS_WITH_STATS_PARSER = ConfigVmsWithStatsParser
    CONFIG_VPX_PATH_2_ID_HOSTS_VMS_PARSER = ConfigVpxPath2IdHostsVmsParser

    def _acceptHeader(self, visitor):
        token, status, vmotion_rate, root, timestamp = self._parseLine()
        assert token == "DrmDomainConfig"

        visitor.visitStatus(status)
        visitor.visitVmotionRate(vmotion_rate)
        visitor.visitRoot(root)
        visitor.visitTimestamp(timestamp)

    def _acceptSections(self, visitor):
        v = visitor.visitHosts()
        self.CONFIG_HOSTS_PARSER(self._fp).accept(v)

        v = visitor.visitVms()
        self.CONFIG_VMS_PARSER(self._fp).accept(v)

        v = visitor.visitGroups()
        self.CONFIG_GROUPS_PARSER(self._fp).accept(v)

        v = visitor.visitHostsWithStats()
        self.CONFIG_HOSTS_WITH_STATS_PARSER(self._fp).accept(v)

        v = visitor.visitVmsWithStats()
        self.CONFIG_VMS_WITH_STATS_PARSER(self._fp).accept(v)

        v = visitor.visitVpxPath2IdHostsVms()
        self.CONFIG_VPX_PATH_2_ID_HOSTS_VMS_PARSER(self._fp).accept(v)

    def accept(self, visitor):

        self._acceptHeader(visitor)
        self._acceptSections(visitor)
        visitor.visitEnd()

class RebalanceModuleParser(TabBasedParser):

    OPTIONS_PARSER = OptionsParser
    SPI_PARSER = SpiParser
    CONFIG_PARSER = ConfigParser
    STATS_PARSER = StatsParser

    def accept(self, visitor):
        _, fairness, migrate = self._parseLine()
        visitor.visitFairnessTimestamp(fairness)
        visitor.visitMigrateTimestamp(migrate)

        v = visitor.visitOptions()
        self.OPTIONS_PARSER(self._fp).accept(v)

        v = visitor.visitSpi()
        self.SPI_PARSER(self._fp).accept(v)

        v = visitor.visitConfig()
        self.CONFIG_PARSER(self._fp).accept(v)

        v = visitor.visitStats()
        self.STATS_PARSER(self._fp).accept(v)

        visitor.visitEnd()

class DumpParser(TabBasedParser):

    REBALANCE_MODULE_PARSER = RebalanceModuleParser
    PROPOSE_ACTIONS_PARSER = ProposeActionsParser

    def __init__(self, fp, *args, **kwargs):
        TabBasedParser.__init__(self, fp, *args, **kwargs)
        self._actions = kwargs.get('actions', False)

    def _acceptRebalanceModule(self, visitor):
        v = visitor.visitRebalanceModule()
        self.REBALANCE_MODULE_PARSER(self._fp).accept(v)

        if self._actions:
            v = visitor.visitProposeActions()
            self.PROPOSE_ACTIONS_PARSER(self._fp).accept(v)

        visitor.visitEnd()

    def accept(self, visitor):
        self._acceptRebalanceModule(visitor)
