# this tool aims at producing valid input for the DRS simulator, based on the
# analysis of a set of DRM dump files

import dummy as base
import sys, os

from functools import partial

class Summary(object):

    def __init__(self):
        self._minTimeStamp = None
        self._maxTimeStamp = None
        self._vms = []
        self._hosts = []
        self._stats = {"VmCpuActAvg5MinPct": [],
                       "VmMemGrantedMB": []}

    @property
    def _aveCpuDemand(self):
        stats = self._stats["VmCpuActAvg5MinPct"]
        return sum(stats)/len(stats)

    @property
    def _aveMemDemand(self):
        stats = self._stats["VmMemGrantedMB"]
        return sum(stats)/len(stats)

    @property
    def _seconds(self):
        if not self._minTimeStamp:
            return 0
        return (self._maxTimeStamp - self._minTimeStamp) / 1000000

    def _recordTimestamp(self, value):
        if self._minTimeStamp is None or value < self._minTimeStamp:
            self._minTimeStamp = value

        if self._maxTimeStamp is None or value > self._maxTimeStamp:
            self._maxTimeStamp = value

    def _receiveStat(self, id, value):
        if self._stats.has_key(id):
            self._stats[id].append(value)

    def __str__(self):
        return (("Total time covered: %d seconds ~= %d minutes ~= %d hours\n"
                 % (self._seconds, self._seconds/60, self._seconds/(60*60))) +
                " minTimeStamp (microseconds): %s\n" % (self._minTimeStamp or "N/A") +
                " maxTimeStamp (microseconds): %s\n" % (self._maxTimeStamp or "N/A") +
                "VM information\n" +
                " Number of VMs: %d\n" % (len(self._vms)) +
                " Ave cpu demand: %d\n" % (self._aveCpuDemand) +
                " Ave mem demand: %d\n" % (self._aveMemDemand) +
                "Host information\n" +
                " Number of Hosts: %d" % (len(self._hosts)))

# FIXME: can we compute something here ?
class DrmOptions(object):

    def __str__(self):
        return "\n".join(["EnablePowerPerformance = 1",
                          "Power = 1",
                          "OverheadGrowth = 4"])

# FIXME:
# - can we compute shares/reservation/limit ?
# - can we compute a RP tree ?
class Resources(object):

    def __init__(self, simvms):
        self._vms = {}
        self._simvms = simvms

    def addVm(self, id, name):
        if name not in self._vms:
            self._vms[id] = name

    def _getVmResources(self, vm):
        name = self._vms[vm]
        return ("vm %s user 1000 0 unlimited 1000 0 unlimited [ /dummy/path/to/%s.cfg ]"
                % (name, name))

    def __str__(self):
        # don't dump the artifacs that DRS produces for unused capacity
        return "\n".join([self._getVmResources(vm) for vm in self._simvms._vmStats.keys()])

class Rules(object):

    def __str__(self):
        return ""

class SimVms(object):

    def __init__(self):
        self._vms = {}
        self._hosts = {}
        self._vmStats = {}
        self._curVm = None

        # this stat has changed names in history...
        self._setInterestingStats("VmMemActiveMBHist")

    def _setInterestingStats(self, memstat):
        self._ACTIVE_MEM_STAT = memstat
        self._interestingStats = ["VmMemSharedZeroMB",
                                  "VmMemSharedNonzeroMB",
                                  "VmMemGrantedMB",
                                  # invariant: this contains the VM demand. See
                                  # DrmSimVm::ResetStats and
                                  # DrmSimHost::SetupVmStats
                                  "VmCpuActAvg5MinPct",
                                  # invariant: this contains the VM demand (the
                                  # simulator caps it, so any greater value
                                  # would provide the same result). See
                                  # DrmSimVm::ResetStats and
                                  # DrmSimVm::UpdateActiveMpn
                                  memstat]

    def addVm(self, id, name, nvcpus, memsize, hostid):
        if not self._vms.has_key(id):
            # we want to register VMs as they are the first time we see them
            # (in particular for the host that runs them)
            self._vms[id] = (name, nvcpus, memsize, hostid)

    def addHost(self, id, name):
        self._hosts[id] = name

    def _getVmName(self, id):
        return "vm%s" % (id)

    def _vmDesc(self, id):
        name, cpus, mem, host = self._vms[id]
        hostname = self._hosts.get(host)
        z_pc, nz_pc = self._vmSharedPct(id)
        # FIXME: too many constants there ?
        wk_period = 0 # should be irrelevant for precomputed and constant anyway
        host_set = "allHosts" # huge assumption, can we do better ?
        return (('%s %d %d precomputed 0 0 constant 300 0 ' +
                 '%d %d precomputed 0 0 constant 300 0 %d %s %s')
                % (self._getVmName(id), cpus, mem, z_pc, nz_pc, wk_period, hostname, host_set))

    def _vmSharedPct(self, vm_id):
        z_pct = []
        nz_pct = []
        for z, nz, g in zip(self._vmStats[vm_id]["VmMemSharedZeroMB"],
                            self._vmStats[vm_id]["VmMemSharedNonzeroMB"],
                            self._vmStats[vm_id][self._ACTIVE_MEM_STAT]):
            if g == 0.0:
                continue
            z_pct.append(z/g)
            nz_pct.append(nz/g)
        if len(z_pct) == 0:
            return (0,0)
        return (int(100*sum(z_pct)/len(z_pct)),
                int(100*sum(nz_pct)/len(nz_pct)))

    def _setVm(self, id):
        if not self._vmStats.has_key(id):
            self._vmStats[id] = dict([(stat, []) for stat in self._interestingStats])
        self._curVm = id

    def _receiveStat(self, id, value):
        if id in self._interestingStats:
            self._vmStats[self._curVm][id].append(value)

    def dumpVmStats(self, dir):
        if not os.path.exists(dir):
            os.mkdir(dir)

        # HACK: recognise that the stats come from a previous simulation on
        # a generated model (first stats don't make sense). Otherwise, insert
        # those meaningless stats.
        # Reason is a confusing behavior of the simulator (maybe a bug?) that
        # casually ignores the first value for cpu model.
        first_cpustats = [self._vmStats[vm]["VmCpuActAvg5MinPct"][0] for vm in self._vmStats.keys()]
        first_memstats = [self._vmStats[vm][self._ACTIVE_MEM_STAT][0] for vm in self._vmStats.keys()]
        if any(first_cpustats) or any(first_memstats):
            # not from a model
            for vm in self._vmStats.keys():
                self._vmStats[vm]["VmCpuActAvg5MinPct"].insert(0, 0)
                self._vmStats[vm][self._ACTIVE_MEM_STAT].insert(0, 0)
        # /HACK

        for vm in self._vmStats.keys():
            # We need to compute a model of the VM that's compatible with the
            # observed statistics.

            # Stats are a function of demand (what the model describes). So for
            # each model dimension we need to find a stat whose function is
            # easily invertible.

            # This does *not* guarantee that the model will be realistic (it
            # doesn't even have to be unique), only that it's as good as it can
            # given the approximations of the simulator algorithm itself.

            name, cpus, mem, host = self._vms[vm]

            f = open(os.path.join(dir, "%s_cpuload.txt" % (self._getVmName(vm))), "w")
            # model gives demand per cpu
            stats = [int(x/cpus) for x in self._vmStats[vm]["VmCpuActAvg5MinPct"]]
            print >> f, "\n".join(map(str, stats))
            f.close()

            f = open(os.path.join(dir, "%s_memload.txt" % (self._getVmName(vm))), "w")
            stats = map(int, self._vmStats[vm][self._ACTIVE_MEM_STAT])
            print >> f, "\n".join(map(str, stats))
            f.close()

    def __str__(self):
        return ("\n".join([self._vmDesc(id) for id in self._vmStats.keys()]))

class SimHosts(object):

    def __init__(self):
        self._hosts = {}

    def addHost(self, id, name, mhz, mem, npcpus, base, cpup, boot, shutdown, state):
        self._hosts[id] = (name, mhz, mem, npcpus, base, cpup, boot, shutdown, state)

    def _hostDesc(self, id):
        # FIXME: certainly we don't always want allDS connectivity...
        return '%s %d %d %d %d %d %d %d %d allDS' % self._hosts[id]

    def __str__(self):
        return ("\n".join([self._hostDesc(id) for id in self._hosts.keys()]))

class Hosts(object):

    def __init__(self):
        self._hosts = []

    def __str__(self):
        return ("\n".join(self._hosts))

    def addHost(self, name):
        if name not in self._hosts:
            self._hosts.append(name)

# some helpers to pass context from visitor to visitor

class _ContextVisitor(object):

    def __init__(self, summary, simvms, simhosts, hosts, options, resources, rules):
        self._summary = summary
        self._simvms = simvms
        self._simhosts = simhosts
        self._hosts = hosts
        self._drmoptions = options
        self._resources = resources
        self._rules = rules

        self._privateInit()

    def _privateInit(self):
        pass

    def _getContextualizedVisitor(self, cls):
        v = cls(self._summary, self._simvms, self._simhosts, self._hosts, self._drmoptions, self._resources, self._rules)
        return v

    @classmethod
    def _makeContextualizedVisitor(cls, c):
        def _meth(this):
            return this._getContextualizedVisitor(c)
        return _meth

def _useContextVisitor(cls):
    return _ContextVisitor._makeContextualizedVisitor(cls)

# now for the real visitors

class VmRecentStatsVisitor(base.RecentStatsVisitor, _ContextVisitor):

    def _privateInit(self):
        self._id = None
        self._stat = None

    def visitTimestamp(self, value):
        self._summary._recordTimestamp(int(value))

    def visitValue(self, value):
        val = float(value)
        # we only care about the latest one, at least for cpu and memory. This
        # might need further adaptation later.
        self._stat = val

    def _setId(self, id):
        self._id = id

    def visitEnd(self):
        self._summary._receiveStat(self._id, self._stat)
        self._simvms._receiveStat(self._id, self._stat)

class VmObjectStatsVisitor(base.ObjectStatsVisitor, _ContextVisitor):

    RECENT_STATS_VISITOR = _useContextVisitor(VmRecentStatsVisitor)

    def visitRecentStats(self, id):
        ret = super(VmObjectStatsVisitor, self).visitRecentStats(id)
        ret._setId(id)
        return ret

class VmStatsVisitor(base.VmStatsVisitor, _ContextVisitor):

    OBJECT_STATS_VISITOR = _useContextVisitor(VmObjectStatsVisitor)

    def visitId(self, value):
        if not value in self._summary._vms:
            self._summary._vms.append(value)
        self._simvms._setVm(value)

class HostStatsVisitor(base.HostStatsVisitor, _ContextVisitor):

    def visitId(self, value):
        if not value in self._summary._hosts:
            self._summary._hosts.append(value)

class StatsVisitor(base.StatsVisitor, _ContextVisitor):

    VM_STATS_VISITOR = _useContextVisitor(VmStatsVisitor)
    HOST_STATS_VISITOR = _useContextVisitor(HostStatsVisitor)

class VmObjectVisitor(base.ObjectVisitor, _ContextVisitor):

    def _privateInit(self):
        self._id = None
        self._name = None

    def visitId(self, value):
        self._id = value

    def visitName(self, value):
        self._name = value

class VmNodeVisitor(base.NodeVisitor, _ContextVisitor):

    OBJECT_VISITOR = _useContextVisitor(VmObjectVisitor)

    def _privateInit(self):
        self._obj = None

    def visitObject(self):
        self._obj = self.OBJECT_VISITOR()
        return self._obj

class VmVisitor(base.VmVisitor, _ContextVisitor):

    NODE_VISITOR = _useContextVisitor(VmNodeVisitor)

    def _privateInit(self):
        self._vcpus = 0
        self._mem = 0
        self._hostid = None
        self._node = None

    def visitRunningOn(self, value):
        self._hostid = value

    def visitNumVcpus(self, value):
        self._vcpus = int(value)

    def visitMem(self, value):
        self._mem = int(value)

    def visitNode(self):
        self._node = self.NODE_VISITOR()
        return self._node

    def _genName(self, id):
        return "vm%s" % (id)

    def visitEnd(self):
        id = self._node._obj._id
        # let's avoid parsing issues with VM names containing spaces...
        # generate a simple name instead of using self._node._obj._name
        name = self._genName(id)
        self._simvms.addVm(id, name, self._vcpus, self._mem, self._hostid)
        self._resources.addVm(id, name)

class VmsVisitor(base.VmsVisitor, _ContextVisitor):

    VM_VISITOR = _useContextVisitor(VmVisitor)

class HostObjectVisitor(base.ObjectVisitor, _ContextVisitor):

    def _privateInit(self):
        self._id = None

    def visitId(self, value):
        self._id = value

class HostVisitor(base.HostVisitor, _ContextVisitor):

    HOST_BASE_POWER = 130
    HOST_CPU_POWER = 55
    OBJECT_VISITOR = _useContextVisitor(HostObjectVisitor)

    def _privateInit(self):
        self._name = None
        self._mhz = 0
        self._mem = 0
        self._cores = 0
        self._boot = 0
        self._shut = 0
        self._state = None

    def visitHostState(self, value):
        self._state = int(value)

    def visitCpuMhz(self, value):
        self._mhz = int(value)

    def visitMemory(self, value):
        self._mem = int(value)

    def visitNumCpuCores(self, value):
        self._cores = int(value)

    def visitBootTime(self, value):
        self._boot = int(value)

    def visitShutDownTime(self, value):
        self._shut = int(value)

    def visitName(self, value):
        self._name = value

    def visitObject(self):
        self._obj = self.OBJECT_VISITOR()
        return self._obj

    def visitCapacity(self, cap, total, reserved=None):
        if cap == "cpu":
            if self._mhz is 0:
                self._mhz = int(total)
                return
        if cap == "mem":
            if self._mem is 0:
                self._mem = int(total)
                return

    def visitEnd(self):
        self._simvms.addHost(self._obj._id, self._name)
        self._simhosts.addHost(self._obj._id, self._name, self._mhz,
                               self._mem, self._cores, self.HOST_BASE_POWER,
                               self.HOST_CPU_POWER, self._boot, self._shut,
                               self._state)
        self._hosts.addHost(self._name)

class HostsVisitor(base.HostsVisitor, _ContextVisitor):

    HOST_VISITOR = _useContextVisitor(HostVisitor)

class ConfigVisitor(base.ConfigVisitor, _ContextVisitor):

    VMS_VISITOR = _useContextVisitor(VmsVisitor)
    HOSTS_VISITOR = _useContextVisitor(HostsVisitor)

class RebalanceModuleVisitor(base.RebalanceModuleVisitor, _ContextVisitor):

    STATS_VISITOR = _useContextVisitor(StatsVisitor)
    CONFIG_VISITOR = _useContextVisitor(ConfigVisitor)

class DumpTool(base.DumpTool, _ContextVisitor):

    def __init__(self, *args, **kwargs):
        self._summary = Summary()
        self._simvms = SimVms()
        self._simhosts = SimHosts()
        self._hosts = Hosts()
        self._drmoptions = DrmOptions()
        self._resources = Resources(self._simvms)
        self._rules = Rules()

        self._output = kwargs.get('output', '.')

    REBALANCE_MODULE_VISITOR = _useContextVisitor(RebalanceModuleVisitor)

    def _dumpObject(self, obj, filename):
        f = open(os.path.join(self._output, filename), "w")
        print >> f, str(obj)
        f.close()

    def visitVersion(self, value):
        if value < 103:
            self._simvms._setInterestingStats("VmMemActiveMB")

    def finalize(self):
        self._dumpObject(self._summary, "summary.txt")
        self._dumpObject(self._simvms, "simvms.txt")
        self._dumpObject(self._simhosts, "simhosts.txt")
        self._dumpObject(self._hosts, "hosts.txt")
        self._dumpObject(self._drmoptions, "drmoptions.txt")
        self._dumpObject(self._resources, "resources.txt")
        self._dumpObject(self._rules, "rules.txt")

        self._simvms.dumpVmStats(os.path.join(self._output, "stats.d"))
